from typing import Union, Optional
import scanpy as sc
import numpy as np
from scanpy.plotting._utils import savefig_or_show
from .. import settings
from ..tools.graph_operations import subset_tree
from .utils import is_categorical, gen_milestones_gradients


def dendrogram(
    adata,
    root_milestone=None,
    milestones=None,
    color_milestones: bool = False,
    color_seg: str = "k",
    linewidth_seg: float = 3,
    alpha_seg: float = 0.3,
    tree_behind: bool = False,
    show_info: bool = True,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs
):
    """\
    Plot the single-cell dendrogram embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    color_milestones
        color the cells with gradients combining pseudotime and milestones.
    color_seg
        color the segments, either a color, or 'seg' colors from obs.seg.
    linewidth_seg
        linewidth of the segments.
    alpha_seg
        alpha of the segments.
    tree_behind
        whether to plot the segment in front or behind the cells.
    show_info
        display the colorbar or not.
    show
        show the plot.
    save
        save the plot.
    kwargs
        arguments to pass to :func:`scanpy.pl.embedding`.

    Returns
    -------
    If `show==False` a tuple of :class:`~matplotlib.axes.Axes`

    """

    if root_milestone is not None:
        verb = settings.verbosity
        settings.verbosity = 1
        adata = subset_tree(adata, root_milestone, milestones, copy=True)
        settings.verbosity = verb

    if color_milestones:
        ax = sc.pl.embedding(adata, basis="dendro", show=False, **kwargs)
        ax.scatter(
            adata.obsm["X_dendro"][:, 0],
            adata.obsm["X_dendro"][:, 1],
            c=gen_milestones_gradients(adata)[adata.obs_names].values,
            s=120000 / adata.shape[0] if "s" not in kwargs else kwargs["s"],
            marker=".",
        )
    else:
        ax = sc.pl.embedding(adata, basis="dendro", show=False, **kwargs)

    for key, value in adata.uns["dendro_segments"].items():
        if "seg_colors" not in adata.uns:
            from . import palette_tools

            palette_tools._set_default_colors_for_categorical_obs(adata, "seg")
        for s in value:
            ax.plot(
                s[0],
                s[1],
                linewidth=linewidth_seg,
                alpha=alpha_seg,
                c=adata.uns["seg_colors"][int(key) - 1]
                if color_seg == "seg"
                else color_seg,
                zorder=0 if tree_behind else None,
            )

    if show_info == False and "color" in kwargs:
        if is_categorical(adata, kwargs["color"]):
            if ax.get_legend() is not None:
                ax.get_legend().remove()
        else:
            # ax.set_box_aspect(aspect=1)
            fig = ax.get_gridspec().figure
            cbar = np.argwhere(
                ["colorbar" in a.get_label() for a in fig.get_axes()]
            ).ravel()
            if len(cbar) > 0:
                fig.get_axes()[cbar[0]].remove()

    savefig_or_show("dendrogram", show=show, save=save)

    if show == False:
        return ax



from pathlib import Path
from typing import Optional
import warnings

import numpy as np
import pandas as pd
from anndata import AnnData

HERE = Path(__file__).parent

from scanpy import read


url_datadir = "https://github.com/LouisFaure/scFates_notebooks/raw/main/"


def morarach20() -> AnnData:
    """\
    10X sequencing from the developping (E15.5) enteric nervous system, it
    includes Schwann Cell precursors and two neuronal population generated
    via a bifurcation.

    Returns
    -------
    Annotated data matrix.
    """

    filename = "data/morarach20.h5ad"
    url = f"{url_datadir}data/morarach20.h5ad"
    return read(filename, backup_url=url, sparse=True, cache=True)


def pancreas() -> AnnData:
    """\
    Data from `Bastidas-Ponce et al. (2019) <https://doi.org/10.1242/dev.173849>`_.
    Pancreatic epithelial and Ngn3-Venus fusion (NVF) cells during secondary transition
    with transcriptome profiles sampled from embryonic day 15.5.

    This datasets has been processed via CellRank and then converted into a scFate tree.

    Returns
    -------
    Annotated data matrix.
    """

    filename = "data/pancreas_quad.h5ad"
    url = f"{url_datadir}data/pancreas_quad.h5ad"
    return read(filename, backup_url=url, sparse=True, cache=True)


def test_adata() -> AnnData:
    """\
    10X sequencing from the developping (E15.5) enteric nervous system, it
    includes Schwann Cell precursors and two neuronal population generated
    via a bifurcation.

    Returns
    -------
    Annotated data matrix.
    """

    filename = HERE / "test.h5ad"
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=FutureWarning, module="anndata")
        return read(filename)



from typing import Optional, Union
from typing_extensions import Literal
from anndata import AnnData
import numpy as np
import pandas as pd
from pandas import DataFrame
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.sparse.csgraph import shortest_path
import igraph
from tqdm import tqdm
import sys
import igraph
import warnings
import itertools
import math
from scipy import sparse
import simpleppt

from ..plot.trajectory import graph as plot_graph
from .. import logging as logg
from .. import settings
from .utils import get_X


def curve(
    adata: AnnData,
    Nodes: int = None,
    use_rep: str = None,
    ndims_rep: Optional[int] = None,
    init: Optional[DataFrame] = None,
    epg_lambda: Optional[Union[float, int]] = 0.01,
    epg_mu: Optional[Union[float, int]] = 0.1,
    epg_trimmingradius: Optional = np.inf,
    epg_initnodes: Optional[int] = 2,
    epg_verbose: bool = False,
    device: Literal["cpu", "gpu"] = "cpu",
    plot: bool = False,
    basis: Optional[str] = "umap",
    seed: Optional[int] = None,
    copy: bool = False,
):
    """\
    Generate a principal curve.

    Learn a curved representation on any space, composed of nodes, approximating the
    position of the cells on a given space such as gene expression, pca, diffusion maps, ...
    Uses ElpiGraph algorithm.

    Parameters
    ----------
    adata
        Annotated data matrix.
    Nodes
        Number of nodes composing the principial tree, use a range of 10 to 100 for
        ElPiGraph approach and 100 to 2000 for PPT approach.
    use_rep
        Choose the space to be learned by the principal tree.
    ndims_rep
        Number of dimensions to use for the inference.
    init
        Initialise the point positions.
    epg_lambda
        Parameter for ElPiGraph, coefficient of ‘stretching’ elasticity [Albergante20]_.
    epg_mu
        Parameter for ElPiGraph, coefficient of ‘bending’ elasticity [Albergante20]_.
    epg_trimmingradius
        Parameter for ElPiGraph, trimming radius for MSE-based data approximation term [Albergante20]_.
    epg_initnodes
        numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
        in the initial step
    epg_verbose
        show verbose output of epg algorithm
    device
        Run method on either `cpu` or on `gpu`
    plot
        Plot the resulting tree.
    basis
        Basis onto which the resulting tree should be projected.
    seed
        A numpy random seed.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['epg']`
            dictionnary containing information from elastic principal curve
        `.obsm['X_R']`
            hard assignment of cells to principal points
        `.uns['graph']['B']`
            adjacency matrix of the principal points
        `.uns['graph']['F']`
            coordinates of principal points in representation space
    """

    logg.info(
        "inferring a principal curve",
        reset=True,
        end=" " if settings.verbosity > 2 else "\n",
    )

    adata = adata.copy() if copy else adata

    if Nodes is None:
        if adata.shape[0] * 2 > 100:
            Nodes = 100
        else:
            Nodes = int(adata.shape[0] / 2)

    logg.hint(
        "parameters used \n"
        "    "
        + str(Nodes)
        + " principal points, mu = "
        + str(epg_mu)
        + ", lambda = "
        + str(epg_lambda)
    )
    curve_epg(
        adata,
        Nodes,
        use_rep,
        ndims_rep,
        init,
        epg_lambda,
        epg_mu,
        epg_trimmingradius,
        epg_initnodes,
        device,
        seed,
        epg_verbose,
    )

    if plot:
        plot_graph(adata, basis)

    return adata if copy else None


def tree(
    adata: AnnData,
    Nodes: int = None,
    use_rep: str = None,
    ndims_rep: Optional[int] = None,
    weight_rep: str = None,
    method: Literal["ppt", "epg"] = "ppt",
    init: Optional[DataFrame] = None,
    ppt_sigma: Optional[Union[float, int]] = 0.1,
    ppt_lambda: Optional[Union[float, int]] = 1,
    ppt_metric: str = "euclidean",
    ppt_nsteps: int = 50,
    ppt_err_cut: float = 5e-3,
    ppt_gpu_tpb: int = 16,
    epg_lambda: Optional[Union[float, int]] = 0.01,
    epg_mu: Optional[Union[float, int]] = 0.1,
    epg_trimmingradius: Optional = np.inf,
    epg_initnodes: Optional[int] = 2,
    epg_verbose: bool = False,
    device: Literal["cpu", "gpu"] = "cpu",
    plot: bool = False,
    basis: Optional[str] = "umap",
    seed: Optional[int] = None,
    copy: bool = False,
):
    """\
    Generate a principal tree.

    Learn a simplified representation on any space, compsed of nodes, approximating the
    position of the cells on a given space such as gene expression, pca, diffusion maps, ...
    If `method=='ppt'`, uses simpleppt implementation from [Soldatov19]_.
    If `method=='epg'`, uses Elastic Principal Graph approach from [Albergante20]_.

    Parameters
    ----------
    adata
        Annotated data matrix.
    Nodes
        Number of nodes composing the principial tree, use a range of 10 to 100 for
        ElPiGraph approach and 100 to 2000 for PPT approach.
    use_rep
        Choose the space to be learned by the principal tree.
    ndims_rep
        Number of dimensions to use for the inference.
    weight_rep
        If `ppt`, use a weight matrix for learning the tree.
    method
        If `ppt`, uses simpleppt approach, `ppt_lambda` and `ppt_sigma` are the
        parameters controlling the algorithm. If `epg`, uses ComputeElasticPrincipalTree
        function from elpigraph python package, `epg_lambda` `epg_mu` and `epg_trimmingradius`
        are the parameters controlling the algorithm.
    init
        Initialise the point positions.
    ppt_sigma
        Regularization parameter for simpleppt [Mao15]_.
    ppt_lambda
        Parameter for simpleppt, penalty for the tree length [Mao15]_.
    ppt_metric
        The metric to use to compute distances in high dimensional space.
        For compatible metrics, check the documentation of
        sklearn.metrics.pairwise_distances if using cpu or
        cuml.metrics.pairwise_distances if using gpu.
    ppt_nsteps
        Number of steps for the optimisation process of simpleppt.
    ppt_err_cut
        Stop simpleppt algorithm if proximity of principal points between iterations less than defiend value.
    ppt_gpu_tpb
        Threads per block parameter for cuda computations.
    epg_lambda
        Parameter for ElPiGraph, coefficient of ‘stretching’ elasticity [Albergante20]_.
    epg_mu
        Parameter for ElPiGraph, coefficient of ‘bending’ elasticity [Albergante20]_.
    epg_trimmingradius
        Parameter for ElPiGraph, trimming radius for MSE-based data approximation term [Albergante20]_.
    epg_initnodes
        numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
        in the initial step
    epg_verbose
        show verbose output of epg algorithm
    device
        Run either mehtod on `cpu` or on `gpu`
    plot
        Plot the resulting tree.
    basis
        Basis onto which the resulting tree should be projected.
    seed
        A numpy random seed.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['ppt']`
            dictionnary containing information from simpelppt tree if method='ppt'
        `.uns['epg']`
            dictionnary containing information from elastic principal tree if method='epg'
        `.obsm['R']`
            soft assignment of cells to principal points
        `.uns['graph']['B']`
            adjacency matrix of the principal points
        `.uns['graph']['F']`
            coordinates of principal points in representation space
    """

    logg.info(
        "inferring a principal tree",
        reset=True,
        end=" " if settings.verbosity > 2 else "\n",
    )

    adata = adata.copy() if copy else adata

    X, use_rep = get_data(adata, use_rep, ndims_rep)

    W = get_data(adata, weight_rep, ndims_rep)[0] if weight_rep is not None else None

    if Nodes is None:
        if adata.shape[0] * 2 > 2000:
            Nodes = 2000
        else:
            Nodes = int(adata.shape[0] / 2)

    if method == "ppt":
        simpleppt.settings.verbosity = settings.verbosity
        ppt = simpleppt.ppt(
            X,
            W,
            Nodes=Nodes,
            init=init,
            sigma=ppt_sigma,
            lam=ppt_lambda,
            metric=ppt_metric,
            nsteps=ppt_nsteps,
            err_cut=ppt_err_cut,
            device=device,
            gpu_tbp=ppt_gpu_tpb,
            seed=seed,
            progress=settings.verbosity > 1,
        )

        ppt = vars(ppt)

        graph = {
            "B": ppt["B"],
            "F": ppt["F"],
            "tips": ppt["tips"],
            "forks": ppt["forks"],
            "metrics": ppt["metric"],
            "use_rep": use_rep,
            "ndims_rep": ndims_rep,
        }

        adata.uns["graph"] = graph
        adata.uns["ppt"] = ppt
        adata.obsm["X_R"] = ppt["R"]

    elif method == "epg":
        graph, epg = tree_epg(
            X,
            Nodes,
            init,
            epg_lambda,
            epg_mu,
            epg_trimmingradius,
            epg_initnodes,
            device,
            seed,
            epg_verbose,
        )
        graph["use_rep"] = use_rep
        graph["ndims_rep"] = ndims_rep
        adata.obsm["X_R"] = graph["R"]
        del graph["R"]
        adata.uns["graph"] = graph
        adata.uns["epg"] = epg

    if plot:
        plot_graph(adata, basis)

    logg.hint(
        "added \n"
        "    .uns['" + method + "'], dictionnary containing inferred tree.\n"
        "    .obsm['X_R'] soft assignment of cells to principal points.\n"
        "    .uns['graph']['B'] adjacency matrix of the principal points.\n"
        "    .uns['graph']['F'] coordinates of principal points in representation space."
    )

    return adata if copy else None


def circle(
    adata: AnnData,
    Nodes: int = None,
    use_rep: str = None,
    ndims_rep: Optional[int] = None,
    init: Optional[DataFrame] = None,
    epg_lambda: Optional[Union[float, int]] = 0.01,
    epg_mu: Optional[Union[float, int]] = 0.1,
    epg_trimmingradius: Optional = np.inf,
    epg_initnodes: Optional[int] = 3,
    epg_verbose: bool = False,
    device: Literal["cpu", "gpu"] = "cpu",
    plot: bool = False,
    basis: Optional[str] = "umap",
    seed: Optional[int] = None,
    copy: bool = False,
):
    """\
    Generate a principal circle.

    Learn a circled representation on any space, composed of nodes, approximating the
    position of the cells on a given space such as gene expression, pca, diffusion maps, ...
    Uses ElpiGraph algorithm.

    Parameters
    ----------
    adata
        Annotated data matrix.
    Nodes
        Number of nodes composing the principial tree, use a range of 10 to 100 for
        ElPiGraph approach and 100 to 2000 for PPT approach.
    use_rep
        Choose the space to be learned by the principal tree.
    ndims_rep
        Number of dimensions to use for the inference.
    init
        Initialise the point positions.
    epg_lambda
        Parameter for ElPiGraph, coefficient of ‘stretching’ elasticity [Albergante20]_.
    epg_mu
        Parameter for ElPiGraph, coefficient of ‘bending’ elasticity [Albergante20]_.
    epg_trimmingradius
        Parameter for ElPiGraph, trimming radius for MSE-based data approximation term [Albergante20]_.
    epg_initnodes
        numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
        in the initial step
    epg_verbose
        show verbose output of epg algorithm
    device
        Run method on either `cpu` or on `gpu`
    plot
        Plot the resulting tree.
    basis
        Basis onto which the resulting tree should be projected.
    seed
        A numpy random seed.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['epg']`
            dictionnary containing information from elastic principal curve
        `.obsm['X_R']`
            soft assignment of cells to principal points
        `.uns['graph']['B']`
            adjacency matrix of the principal points
        `.uns['graph']['F']`
            coordinates of principal points in representation space
    """

    logg.info(
        "inferring a principal circle",
        reset=True,
        end=" " if settings.verbosity > 2 else "\n",
    )

    adata = adata.copy() if copy else adata

    if Nodes is None:
        if adata.shape[0] * 2 > 100:
            Nodes = 100
        else:
            Nodes = int(adata.shape[0] / 2)

    logg.hint(
        "parameters used \n"
        "    "
        + str(Nodes)
        + " principal points, mu = "
        + str(epg_mu)
        + ", lambda = "
        + str(epg_lambda)
    )
    circle_epg(
        adata,
        Nodes,
        use_rep,
        ndims_rep,
        init,
        epg_lambda,
        epg_mu,
        epg_trimmingradius,
        epg_initnodes,
        device,
        seed,
        epg_verbose,
    )

    if plot:
        plot_graph(adata, basis)

    return adata if copy else None


def tree_epg(
    X,
    Nodes: int = None,
    init: Optional[DataFrame] = None,
    lam: Optional[Union[float, int]] = 0.01,
    mu: Optional[Union[float, int]] = 0.1,
    trimmingradius: Optional = np.inf,
    initnodes: int = None,
    device: str = "cpu",
    seed: Optional[int] = None,
    verbose: bool = True,
):

    try:
        import elpigraph

    except Exception as e:
        warnings.warn(
            'ElPiGraph package is not installed \
            \nPlease use "pip install git+https://github.com/j-bac/elpigraph-python.git" to install it'
        )
    logg.hint(
        "parameters used \n"
        "    "
        + str(Nodes)
        + " principal points, mu = "
        + str(mu)
        + ", lambda = "
        + str(lam)
    )

    if seed is not None:
        np.random.seed(seed)

    if device == "gpu":
        import cupy as cp
        from cuml.metrics import pairwise_distances
        from .utils import cor_mat_gpu

        Tree = elpigraph.computeElasticPrincipalTree(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            GPU=True,
            verbose=verbose,
        )

        R = pairwise_distances(
            cp.asarray(X.values), cp.asarray(Tree[0]["NodePositions"])
        )

        R = cp.asnumpy(R)
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    else:
        from .utils import cor_mat_cpu
        from sklearn.metrics import pairwise_distances

        Tree = elpigraph.computeElasticPrincipalTree(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            verbose=verbose,
        )

        R = pairwise_distances(X.values, Tree[0]["NodePositions"])
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    g = igraph.Graph(directed=False)
    g.add_vertices(np.unique(Tree[0]["Edges"][0].flatten().astype(int)))
    g.add_edges(
        pd.DataFrame(Tree[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
    )

    # mat = np.asarray(g.get_adjacency().data)
    # mat = mat + mat.T - np.diag(np.diag(mat))
    # B=((mat>0).astype(int))

    B = np.asarray(g.get_adjacency().data)

    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
    B = B[sel, :][:, sel]
    R = R[:, sel]
    F = Tree[0]["NodePositions"].T[:, sel]
    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
    tips = np.argwhere(np.array(g.degree()) == 1).flatten()

    def reconnect():
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        distmat = np.triu(pairwise_distances(F[:, tips].T))
        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
        distmat[distmat == 0] = np.inf
        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
        i, j = distmat.index[row], distmat.columns[col]
        B[i, j] = 1
        B[j, i] = 1
        return B

    if len(emptynodes) > 0:
        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))

    recon = len(np.unique(np.array(g.clusters().membership))) > 1
    while recon:
        B = reconnect()
        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        recon = len(np.unique(np.array(g.clusters().membership))) > 1

    forks = np.argwhere(np.array(g.degree()) > 2).flatten()

    graph = {
        "B": B,
        "R": R,
        "F": Tree[0]["NodePositions"].T,
        "tips": tips,
        "forks": forks,
        "cells_fitted": X.index.tolist(),
        "metrics": "euclidean",
    }

    Tree[0]["Edges"] = list(Tree[0]["Edges"])[0]

    return graph, Tree[0]


def curve_epg(
    adata: AnnData,
    Nodes: int = None,
    use_rep: str = None,
    ndims_rep: Optional[int] = None,
    init: Optional[DataFrame] = None,
    lam: Optional[Union[float, int]] = 0.01,
    mu: Optional[Union[float, int]] = 0.1,
    trimmingradius: Optional = np.inf,
    initnodes: int = None,
    device: str = "cpu",
    seed: Optional[int] = None,
    verbose: bool = True,
):
    try:
        import elpigraph

    except Exception as e:
        warnings.warn(
            'ElPiGraph package is not installed \
            \nPlease use "pip install git+https://github.com/j-bac/elpigraph-python.git" to install it'
        )

    X, use_rep = get_data(adata, use_rep, ndims_rep)

    if seed is not None:
        np.random.seed(seed)

    if device == "gpu":
        import cupy as cp
        from .utils import cor_mat_gpu
        from cuml.metrics import pairwise_distances

        Curve = elpigraph.computeElasticPrincipalCurve(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            GPU=True,
            verbose=verbose,
        )

        R = pairwise_distances(
            cp.asarray(X.values), cp.asarray(Curve[0]["NodePositions"])
        )

        R = cp.asnumpy(R)
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    else:
        from .utils import cor_mat_cpu
        from sklearn.metrics import pairwise_distances

        Curve = elpigraph.computeElasticPrincipalCurve(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            verbose=verbose,
        )

        R = pairwise_distances(X.values, Curve[0]["NodePositions"])
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    g = igraph.Graph(directed=False)
    g.add_vertices(np.unique(Curve[0]["Edges"][0].flatten().astype(int)))
    g.add_edges(
        pd.DataFrame(Curve[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
    )

    # mat = np.asarray(g.get_adjacency().data)
    # mat = mat + mat.T - np.diag(np.diag(mat))
    # B=((mat>0).astype(int))

    B = np.asarray(g.get_adjacency().data)

    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
    B = B[sel, :][:, sel]
    R = R[:, sel]
    F = Curve[0]["NodePositions"].T[:, sel]
    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
    tips = np.argwhere(np.array(g.degree()) == 1).flatten()

    def reconnect():
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        distmat = np.triu(pairwise_distances(F[:, tips].T))
        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
        distmat[distmat == 0] = np.inf
        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
        i, j = distmat.index[row], distmat.columns[col]
        B[i, j] = 1
        B[j, i] = 1
        return B

    if len(emptynodes) > 0:
        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))

    recon = len(np.unique(np.array(g.clusters().membership))) > 1
    while recon:
        B = reconnect()
        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        recon = len(np.unique(np.array(g.clusters().membership))) > 1

    forks = np.argwhere(np.array(g.degree()) > 2).flatten()

    graph = {
        "B": B,
        "F": Curve[0]["NodePositions"].T,
        "tips": tips,
        "forks": forks,
        "metrics": "euclidean",
        "use_rep": use_rep,
        "ndims_rep": ndims_rep,
    }

    Curve[0]["Edges"] = list(Curve[0]["Edges"])[0]

    adata.uns["graph"] = graph
    adata.uns["epg"] = Curve[0]
    adata.obsm["X_R"] = R

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .uns['epg'] dictionnary containing inferred elastic curve generated from elpigraph.\n"
        "    .obsm['X_R'] hard assignment of cells to principal points.\n"
        "    .uns['graph']['B'] adjacency matrix of the principal points.\n"
        "    .uns['graph']['F'], coordinates of principal points in representation space."
    )

    return adata


def circle_epg(
    adata: AnnData,
    Nodes: int = None,
    use_rep: str = None,
    ndims_rep: Optional[int] = None,
    init: Optional[DataFrame] = None,
    lam: Optional[Union[float, int]] = 0.01,
    mu: Optional[Union[float, int]] = 0.1,
    trimmingradius: Optional = np.inf,
    initnodes: int = None,
    device: str = "cpu",
    seed: Optional[int] = None,
    verbose: bool = True,
):
    try:
        import elpigraph

    except Exception as e:
        warnings.warn(
            'ElPiGraph package is not installed \
            \nPlease use "pip install git+https://github.com/j-bac/elpigraph-python.git" to install it'
        )

    X, use_rep = get_data(adata, use_rep, ndims_rep)

    if seed is not None:
        np.random.seed(seed)

    if device == "gpu":
        import cupy as cp
        from .utils import cor_mat_gpu
        from cuml.metrics import pairwise_distances

        Curve = elpigraph.computeElasticPrincipalCircle(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            GPU=True,
            verbose=verbose,
        )

        R = pairwise_distances(
            cp.asarray(X.values), cp.asarray(Curve[0]["NodePositions"])
        )

        R = cp.asnumpy(R)
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    else:
        from .utils import cor_mat_cpu
        from sklearn.metrics import pairwise_distances

        Curve = elpigraph.computeElasticPrincipalCircle(
            X.values.astype(np.float64),
            NumNodes=Nodes,
            Do_PCA=False,
            InitNodes=initnodes,
            Lambda=lam,
            Mu=mu,
            TrimmingRadius=trimmingradius,
            verbose=verbose,
        )

        R = pairwise_distances(X.values, Curve[0]["NodePositions"])
        # Hard assigment
        R = sparse.csr_matrix(
            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
        ).A

    g = igraph.Graph(directed=False)
    g.add_vertices(np.unique(Curve[0]["Edges"][0].flatten().astype(int)))
    g.add_edges(
        pd.DataFrame(Curve[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
    )

    # mat = np.asarray(g.get_adjacency().data)
    # mat = mat + mat.T - np.diag(np.diag(mat))
    # B=((mat>0).astype(int))

    B = np.asarray(g.get_adjacency().data)

    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
    B = B[sel, :][:, sel]
    R = R[:, sel]
    F = Curve[0]["NodePositions"].T[:, sel]
    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
    tips = np.argwhere(np.array(g.degree()) == 1).flatten()

    def reconnect():
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        distmat = np.triu(pairwise_distances(F[:, tips].T))
        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
        distmat[distmat == 0] = np.inf
        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
        i, j = distmat.index[row], distmat.columns[col]
        B[i, j] = 1
        B[j, i] = 1
        return B

    if len(emptynodes) > 0:
        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))

    recon = len(tips) > 0
    while recon:
        B = reconnect()
        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
        recon = len(tips) > 0

    forks = np.argwhere(np.array(g.degree()) > 2).flatten()

    graph = {
        "B": B,
        "F": F,
        "tips": tips,
        "forks": forks,
        "metrics": "euclidean",
        "use_rep": use_rep,
        "ndims_rep": ndims_rep,
    }

    Curve[0]["Edges"] = list(Curve[0]["Edges"])[0]

    adata.uns["graph"] = graph
    adata.uns["epg"] = Curve[0]
    adata.obsm["X_R"] = R

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .uns['epg'] dictionnary containing inferred elastic circle generated from elpigraph.\n"
        "    .obsm['X_R'] hard assignment of cells to principal points.\n"
        "    .uns['graph']['B'] adjacency matrix of the principal points.\n"
        "    .uns['graph']['F'], coordinates of principal points in representation space."
    )

    return adata


def get_data(adata, use_rep, ndims_rep):

    if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
        use_rep = f"X_{use_rep}"

    if (
        (use_rep not in adata.layers.keys())
        & (use_rep not in adata.obsm.keys())
        & (use_rep != "X")
    ):
        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
        n_pcs = None if use_rep == "X" else n_pcs

    if use_rep == "X":
        ndims_rep = None
        if sparse.issparse(adata.X):
            X = DataFrame(adata.X.A, index=adata.obs_names)
        else:
            X = DataFrame(adata.X, index=adata.obs_names)
    elif use_rep in adata.layers.keys():
        if sparse.issparse(adata.layers[use_rep]):
            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
        else:
            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
    elif use_rep in adata.obsm.keys():
        X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)

    if ndims_rep is not None:
        X = X.iloc[:, :ndims_rep]

    return X, use_rep



import numpy as np
import pandas as pd
from anndata import AnnData

import igraph
import matplotlib.pyplot as plt
from scipy import sparse

from typing import Union, Optional
from typing_extensions import Literal
from scanpy.plotting._utils import savefig_or_show
from ..tools.utils import getpath
from .trajectory import trajectory as plot_trajectory
from .utils import setup_axes
from ..tools.graph_operations import subset_tree
from ..get import modules as get_modules
from .. import settings

import scanpy as sc


def modules(
    adata: AnnData,
    root_milestone,
    milestones,
    color: str = "milestones",
    module: Literal["early", "late", "all"] = "all",
    show_traj: bool = False,
    layer: Optional[str] = None,
    smooth: bool = False,
    ax_early=None,
    ax_late=None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs,
):
    """\
    Plot the mean expression of the early and late modules.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    color
        color the cells with variable from adata.obs.
    module
        whether to show early, late or both modules.
    show_traj
        show trajectory on the early module plot.
    layer
        layer to use to compute mean of module.
    smooth
        whether to smooth the data using knn graph.
    ax_early
        existing axes for early module.
    ax_late
        existing axes for late module.
    show
        show the plot.
    save
        save the plot.
    kwargs
        arguments to pass to :func:`scFates.pl.trajectory` if `show_traj=True`, else to :func:`scanpy.pl.embedding`

    Returns
    -------
    If `show==False` a tuple of :class:`~matplotlib.axes.Axes`

    """

    plt.rcParams["axes.grid"] = False

    X_modules = get_modules(adata, root_milestone, milestones, layer)

    cells = X_modules.index
    X_early, X_late = X_modules.iloc[:, :2], X_modules.iloc[:, 2:]
    verb = settings.verbosity
    settings.verbosity = 1
    nmil = len(adata.uns["graph"]["milestones"])
    if nmil > 4:
        adata_c = subset_tree(adata, root_milestone, milestones, copy=True)
        adata_c.obsm["X_early"] = X_early.loc[adata_c.obs_names].values
        adata_c.obsm["X_late"] = X_late.loc[adata_c.obs_names].values
    else:
        adata_c = AnnData(
            X_early.values,
            obs=adata.obs,
            uns=adata.uns,
            obsm={
                "X_early": X_early.values,
                "X_late": X_late.values,
                "X_R": adata.obsm["X_R"],
            },
            obsp=adata.obsp,
        )
    settings.verbosity = verb

    if smooth:
        adata_c.obsm["X_early"] = adata_c.obsp["connectivities"].dot(
            adata_c.obsm["X_early"]
        )
        adata_c.obsm["X_late"] = adata_c.obsp["connectivities"].dot(
            adata_c.obsm["X_late"]
        )

    if module == "all":
        if (ax_early is None) & (ax_late is None):
            axs, _, _, _ = setup_axes(panels=[0, 1])
            ax_early, ax_late = axs
    elif module == "early":
        if ax_early is None:
            axs, _, _, _ = setup_axes(panels=[0])
            ax_early = axs[0]
    elif module == "late":
        if ax_late is None:
            axs, _, _, _ = setup_axes(panels=[0])
            ax_late = axs[0]

    color = "old_milestones" if ((color == "milestones") & (nmil > 4)) else color

    if (module == "early") | (module == "all"):
        if show_traj:
            plot_trajectory(
                adata_c,
                basis="early",
                root_milestone=root_milestone,
                milestones=milestones,
                color_cells=color,
                show=False,
                title="",
                legend_loc="none",
                ax=ax_early,
                **kwargs,
            )
        else:
            sc.pl.embedding(
                adata_c[cells],
                basis="early",
                color=color,
                legend_loc="none",
                title="",
                show=False,
                ax=ax_early,
                **kwargs,
            )
        ax_early.set_xlabel("early " + milestones[0])
        ax_early.set_ylabel("early " + milestones[1])

    if (module == "late") | (module == "all"):
        sc.pl.embedding(
            adata_c[cells],
            basis="late",
            color=color,
            legend_loc="none",
            show=False,
            title="",
            ax=ax_late,
            **kwargs,
        )
        ax_late.set_xlabel("late " + milestones[0])
        ax_late.set_ylabel("late " + milestones[1])

    savefig_or_show("modules", show=show, save=save)

    if show == False:
        if module == "all":
            return ax_early, ax_late
        elif module == "early":
            return ax_early
        elif module == "late":
            return ax_late



import os

os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"

from typing import Union, Optional, Iterable

import numpy as np
import pandas as pd
from functools import partial
from anndata import AnnData
import shutil
import sys
import copy
import igraph
import warnings
from functools import reduce
from statsmodels.stats.multitest import multipletests
import statsmodels.formula.api as sm

from joblib import delayed
from tqdm import tqdm

from .. import logging as logg
from .. import settings
from .utils import getpath, ProgressParallel, get_X, importeR

Rpy2, R, rstats, rmgcv, Formula = importeR("performing bifurcation analysis")
check = [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]


def test_fork(
    adata: AnnData,
    root_milestone,
    milestones,
    features: Union[None, str] = None,
    rescale: bool = False,
    layer: Optional[str] = None,
    n_jobs: int = 1,
    n_map: int = 1,
    copy: bool = False,
):

    """\
    Test for branch differential gene expression and differential upregulation from progenitor to terminal state.

    First, differential gene expression between two branches is performed. This is performed by
    comparing the full model :math:`g_{i} \\sim\ pseudotime*Branch` to the reduced one
    :math:`g_{i} \\sim\ pseudotime` using F-test.

    Then, each feature is tested for its upregulation along the path from progenitor to terminal state,
    using the linear model :math:`g_{i} \sim\ pseudotime`.


    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    features
        Which features to test (all by default).
    rescale
        By default, analysis restrict to only cells having a pseudotime lower than the shortest branch maximum pseudotime, this can be avoided by rescaling the post bifurcation pseudotime of both branches to 1.
    layer
        layer to use for the test
    n_map
        number of cell mappings from which to do the test.
    n_jobs
        number of cpu processes used to perform the test.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['fork']`
            DataFrame with fork test results.

    """

    if any(check):
        idx = np.argwhere(
            [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]
        ).min()
        raise Exception(np.array([Rpy2, R, rstats, rmgcv, Formula])[idx])

    adata = adata.copy() if copy else adata

    genes = adata.var_names if features is None else features

    logg.info("testing fork", reset=True)

    graph = adata.uns["graph"]

    uns_temp = adata.uns.copy()

    dct = graph["milestones"]
    dct_rev = dict(zip(dct.values(), dct.keys()))

    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    check_t = [graph["root"]]
    if "root2" in graph:
        check_t = [*check_t, graph["root2"]]
    check_t = [*check_t, *graph["forks"]]

    if root not in check_t:
        raise Exception("The root milestone chosen is neither the root nor a fork!")

    g = igraph.Graph.Adjacency((graph["B"] > 0).tolist(), mode="undirected")
    # Add edge weights and node labels.
    g.es["weight"] = graph["B"][graph["B"].nonzero()]

    vpath = g.get_shortest_paths(root, leaves)
    interPP = list(set(vpath[0]) & set(vpath[1]))
    vpath = g.get_shortest_paths(graph["pp_info"].loc[interPP, :].time.idxmax(), leaves)

    fork_stat = list()
    upreg_stat = list()

    for m in tqdm(
        range(n_map), disable=n_map == 1, file=sys.stdout, desc="    multi mapping "
    ):

        if n_map == 1:
            logg.info("    single mapping")
        ## Diff expr between forks

        df = adata.uns["pseudotime_list"][str(m)]
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        def get_branches(m):
            leave = adata.uns["graph"]["milestones"][m]
            ddf = getpath(img, root, adata.uns["graph"]["tips"], leave, graph, df)
            ddf["i"] = adata.uns["graph"]["milestones"][m]
            return ddf

        brcells = pd.concat([get_branches(m) for m in milestones])
        brcells = brcells.loc[brcells.index.drop_duplicates(False)]

        matw = None
        if matw is None:
            brcells["w"] = 1
        else:
            brcells["w"] = matw[gene, :][:, graph["cells_fitted"]]

        brcells.drop(["seg", "edge"], axis=1, inplace=True)

        if rescale:
            for i in brcells.i.unique():
                brcells.loc[brcells.i == i, "t"] = (
                    brcells.loc[brcells.i == i].t - brcells.loc[brcells.i == i].t.min()
                ) / (
                    brcells.loc[brcells.i == i].t.max()
                    - brcells.loc[brcells.i == i].t.min()
                )

        Xgenes = get_X(adata, brcells.index, genes, layer, togenelist=True)

        data = list(zip([brcells] * len(Xgenes), Xgenes))

        stat = ProgressParallel(
            n_jobs=n_jobs,
            use_tqdm=n_map == 1,
            total=len(data),
            file=sys.stdout,
            desc="    Differential expression",
        )(delayed(gt_fun)(data[d]) for d in range(len(data)))

        stat = pd.concat(stat, axis=1).T
        stat.index = genes
        stat.columns = [dct_rev[c] for c in stat.columns[: len(milestones)]] + ["de_p"]

        fork_stat = fork_stat + [stat]

        topleave = fork_stat[m].iloc[:, :-1].idxmax(axis=1).apply(lambda mil: dct[mil])

        ## test for upregulation
        if n_map == 1:
            logg.info("    test for upregulation for each leave vs root")
        leaves_stat = list()
        for leave in leaves:
            subtree = getpath(img, root, graph["tips"], leave, graph, df).sort_values(
                "t"
            )

            topgenes = topleave[topleave == leave].index

            Xgenes = get_X(adata, subtree.index, topgenes, layer, togenelist=True)

            data = list(zip([subtree] * len(Xgenes), Xgenes))

            stat = ProgressParallel(
                n_jobs=n_jobs,
                use_tqdm=n_map == 1,
                total=len(data),
                file=sys.stdout,
                desc="    upreg " + str(keys[vals == leave][0]),
            )(delayed(test_upreg)(data[d]) for d in range(len(data)))

            stat = pd.DataFrame(stat, index=topgenes, columns=["up_A", "up_p"])
            leaves_stat = leaves_stat + [stat]

        upreg_stat = upreg_stat + [pd.concat(leaves_stat).loc[genes]]

    # summarize fork statistics
    # fork_stat=list(map(lambda x: pd.DataFrame(x,index=genes,columns=["effect","p_val"]),fork_stat))

    fdr_l = list(
        map(
            lambda x: pd.Series(
                multipletests(x.de_p, method="bonferroni")[1], index=x.index, name="fdr"
            ),
            fork_stat,
        )
    )

    st_l = list(
        map(
            lambda x: pd.Series(
                (x.de_p < 5e-2).values * 1, index=x.index, name="signi_p"
            ),
            fork_stat,
        )
    )
    stf_l = list(
        map(
            lambda x: pd.Series((x < 5e-2).values * 1, index=x.index, name="signi_fdr"),
            fdr_l,
        )
    )

    fork_stat = list(
        map(
            lambda w, x, y, z: pd.concat([w, x, y, z], axis=1),
            fork_stat,
            fdr_l,
            st_l,
            stf_l,
        )
    )

    effect = list(
        map(
            lambda i: pd.concat(
                list(map(lambda x: x.iloc[:, i], fork_stat)), axis=1
            ).median(axis=1),
            range(len(milestones)),
        )
    )
    p_val = pd.concat(list(map(lambda x: x.de_p, fork_stat)), axis=1).median(axis=1)
    fdr = pd.concat(list(map(lambda x: x.fdr, fork_stat)), axis=1).median(axis=1)
    signi_p = pd.concat(list(map(lambda x: x.signi_p, fork_stat)), axis=1).mean(axis=1)
    signi_fdr = pd.concat(list(map(lambda x: x.signi_fdr, fork_stat)), axis=1).mean(
        axis=1
    )

    colnames = fork_stat[0].columns
    fork_stat = pd.concat(effect + [p_val, fdr, signi_p, signi_fdr], axis=1)
    fork_stat.columns = colnames

    # summarize upregulation stats
    colnames = upreg_stat[0].columns
    upreg_stat = pd.concat(
        list(
            map(
                lambda i: pd.concat(
                    list(map(lambda x: x.iloc[:, i], upreg_stat)), axis=1
                ).median(axis=1),
                range(2),
            )
        ),
        axis=1,
    )
    upreg_stat.columns = colnames

    summary_stat = pd.concat([fork_stat, upreg_stat], axis=1)
    adata.uns = uns_temp

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    name = root_milestone + "->" + "<>".join(milestones)

    # adata.uns[name]["fork"] = summary_stat

    adata.uns[name] = {"fork": summary_stat}
    logg.hint(
        "added \n" "    .uns['" + name + "']['fork'], DataFrame with fork test results."
    )

    return adata if copy else None


def gt_fun(data):

    sdf = data[0]
    sdf["exp"] = data[1]

    ## add weighted matrix part
    #
    #

    global rmgcv

    m = rmgcv.gam(
        Formula("exp ~ s(t)+s(t,by=as.factor(i))+as.factor(i)"),
        data=sdf,
        weights=sdf["w"],
    )

    tmin = np.min([sdf.loc[sdf.i == i, "t"].max() for i in sdf.i.unique()])
    sdf = sdf.loc[sdf.t <= tmin]
    Amps = sdf.groupby("i").apply(lambda x: np.mean(x.exp))
    res = Amps - Amps.max()
    res["de_p"] = rmgcv.summary_gam(m)[3][1]

    return res


def test_upreg(data):

    sdf = data[0]
    sdf["exp"] = data[1]

    result = sm.ols(formula="exp ~ t", data=sdf).fit()
    return [result.params["t"], result.pvalues["t"]]


def branch_specific(
    adata: AnnData,
    root_milestone,
    milestones,
    effect: float = None,
    stf_cut: float = 0.7,
    up_A: float = 0,
    up_p: float = 5e-2,
    copy: bool = False,
):

    """\
    Assign genes differentially expressed between two post-bifurcation branches.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    effect
        minimum expression differences to call gene as differentially upregulated.
    stf_cut
        fraction of projections when gene passed fdr < 0.05.
    up_A
        minimum expression increase at derivative compared to progenitor branches to call gene as branch-specific.
    up_p
        p-value of expression changes of derivative compared to progenitor branches to call gene as branch-specific.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['branch']`
            assigned branch.

    """

    adata = adata.copy() if copy else adata

    uns_temp = adata.uns.copy()

    name = root_milestone + "->" + "<>".join(milestones)

    df = adata.uns[name]["fork"]

    df = df[(df.up_A > up_A) & (df.up_p < up_p) & (df.signi_fdr > stf_cut)]
    df = df[((df.iloc[:, : len(milestones)] + effect) > 0).sum(axis=1) == 1]
    df["branch"] = df.iloc[:, : len(milestones)].idxmax(axis=1)

    logg.info(
        "    "
        + "branch specific features: "
        + ", ".join(
            [
                ": ".join(st)
                for st in list(
                    zip(
                        df.value_counts("branch").index,
                        df.value_counts("branch").values.astype(str),
                    )
                )
            ]
        )
    )

    adata.uns = uns_temp

    adata.uns[name]["fork"] = df

    logg.info("    finished", time=False, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "updated \n"
        "    .uns['"
        + name
        + "']['fork'], DataFrame updated with additionnal 'branch' column."
    )

    return adata if copy else None


def activation(
    adata: AnnData,
    root_milestone,
    milestones,
    deriv_cut: float = 0.15,
    pseudotime_offset: float = 0,
    nwin: int = 20,
    steps: int = 5,
    n_map: int = 1,
    copy: bool = False,
    n_jobs=-1,
    layer=None,
):

    """\
    Identify pseudotime of activation of branch-specififc features.

    This aims in classifying the genes according to their their activation timing
    compared to the pseudotime of the bifurcation. Any feature activated before the
    bifurcation is considered as 'early', others are considered 'late'.

    This is done by separating the path into bins of equal pseudotime, and then
    identifying successive bins having a change in expression higher than the
    parameter `deriv_cut`.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    deriv_cut
        a first passage of derivative at this cutoff (in proportion to the full dynamic range of the fitted feature) is considered as activation timing
    pseudotime_offset
        consider a feature as early if it gets activated before: pseudotime at bifurcation-pseudotime_offset.
    nwin
        windows of pseudotime to use for assessing activation timimg
    steps
        number of steps dividing a window for that will slide along the pseudotime
    n_map
        number of cell mappings from which to do the test.
    n_jobs
        number of cpu processes used to perform the test.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['module']`
            classify feature as 'early' or 'late'.
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['activation']`
            pseudotime of activationh.
    """

    if any(check):
        idx = np.argwhere(
            [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]
        ).min()
        raise Exception(np.array([Rpy2, R, rstats, rmgcv, Formula])[idx])

    adata = adata.copy() if copy else adata

    graph = adata.uns["graph"]

    logg.info("testing activation", reset=True)

    uns_temp = adata.uns.copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)
    stats = adata.uns[name]["fork"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    allact = []

    def activation_map(m):
        df = adata.uns["pseudotime_list"][str(m)]
        acti = pd.Series(0, index=stats.index)

        for leave in leaves:
            subtree = getpath(img, root, graph["tips"], leave, graph, df).sort_values(
                "t"
            )
            del subtree["branch"]
            subtree["deriv_cut"] = deriv_cut
            subtree["nwin"] = nwin
            subtree["steps"] = steps

            genes = stats.index[stats["branch"] == str(keys[vals == leave][0])]

            wf = warnings.filters.copy()
            warnings.filterwarnings("ignore")

            Xgenes = get_X(adata, subtree.index, genes, layer, togenelist=True)

            warnings.filters = wf

            data = list(zip([subtree] * len(Xgenes), Xgenes))

            acti.loc[genes] = ProgressParallel(
                n_jobs=n_jobs if n_map == 1 else 1,
                total=len(data),
                use_tqdm=n_map == 1,
                file=sys.stdout,
                desc="    to " + str(keys[vals == leave][0]),
            )(delayed(get_activation)(data[d]) for d in range(len(data)))

        return acti

    allact = ProgressParallel(
        n_jobs=n_jobs if n_map > 1 else 1,
        total=n_map,
        use_tqdm=n_map > 1,
        file=sys.stdout,
        desc="    mapping: ",
    )(delayed(activation_map)(m) for m in range(n_map))

    stats["activation"] = pd.concat(allact, axis=1).median(axis=1).values

    common_seg = list(
        set.intersection(
            *list(
                map(lambda l: set(img.get_shortest_paths(str(root), str(l))[0]), leaves)
            )
        )
    )
    common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
    fork_t = (
        adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max() - pseudotime_offset
    )

    logg.info("    threshold pseudotime is: " + str(fork_t))

    stats["module"] = "early"
    stats.loc[stats["activation"] > fork_t, "module"] = "late"

    adata.uns = uns_temp

    adata.uns[name]["fork"] = stats

    for l in leaves:
        c_early = np.sum(
            (stats.branch == str(keys[vals == l][0])) & (stats.module == "early")
        )
        c_late = np.sum(
            (stats.branch == str(keys[vals == l][0])) & (stats.module == "late")
        )
        logg.info(
            "    "
            + str(c_early)
            + " early and "
            + str(c_late)
            + " late features specific to leave "
            + str(keys[vals == l][0])
        )

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "updated \n"
        "    .uns['"
        + name
        + "']['fork'], DataFrame updated with additionnal 'activation' and 'module' columns."
    )

    return adata if copy else None


def get_activation(data):
    global rmgcv
    subtree = data[0]
    subtree["exp"] = data[1]
    # load parameters
    deriv_cut = subtree["deriv_cut"][0]
    nwin = subtree["nwin"][0]
    steps = subtree["steps"][0]

    wf = warnings.filters.copy()
    warnings.filterwarnings("ignore")

    def gamfit(sdf):
        m = rmgcv.gam(Formula("exp ~ s(t)"), data=sdf, gamma=1)
        return rmgcv.predict_gam(m)

    subtree["fitted"] = gamfit(subtree)

    window = (subtree.t.max() - subtree.t.min()) / nwin
    step = window / steps

    wins = pd.DataFrame(
        {
            "start": np.array([step * i for i in range(nwin * steps)])
            + subtree.t.min(),
            "end": np.array([step * i for i in range(nwin * steps)])
            + subtree.t.min()
            + window,
        }
    )
    wins = wins.loc[wins.end < subtree.t.max()]

    df_t = subtree.t.sort_values()

    fitted = subtree.fitted
    fitted = (fitted - fitted.min()) / (fitted.max() - fitted.min())
    fitted = fitted[df_t.index]
    changes = wins.apply(
        lambda x: (fitted.loc[df_t[(df_t > x[0]) & (df_t < x[1])].index]).diff().sum(),
        axis=1,
    )

    if sum(changes > deriv_cut) == 0:
        act = subtree.t.max() + 1
    else:
        act = wins.mean(axis=1)[changes > deriv_cut].min()
    warnings.filters = wf

    return act


def activation_lm(
    adata: AnnData,
    root_milestone,
    milestones,
    fdr_cut: float = 0.05,
    stf_cut: float = 0.8,
    pseudotime_offset: float = 0,
    n_map: int = 1,
    copy: bool = False,
    n_jobs=-1,
    layer=None,
):

    """\
    A more robust version of `tl.activation`.

    This is considered to be a more robust version of :func:`scFates.tl.activation`.
    The common path between the two fates is retained for analysis, each feature is
    tested for its upregulation along the path from progenitor to the fork,
    using the linear model :math:`g_{i} \sim\ pseudotime`.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    stf_cut
        fraction of projections when gene passed fdr < 0.05.
    pseudotime_offset
        consider cells to retain up to the pseudotime_fork-pseudotime_offset.
    n_map
        number of cell mappings from which to do the test.
    n_jobs
        number of cpu processes used to perform the test.
    copy
        Return a copy instead of writing to adata.
    layer
        layer to use for the test

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['module']`
            classify feature as 'early' or 'late'.
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['slope']`
            slope calculated by the linear model.
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['pval']`
            pval resulting from linear model.
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['fdr']`
            corrected fdr value.
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']['prefork_signi']`
            proportion of projections where fdr<0.05.

    """

    graph = adata.uns["graph"]

    uns_temp = adata.uns.copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)
    stats = adata.uns[name]["fork"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    common_seg = list(
        set.intersection(
            *list(
                map(lambda l: set(img.get_shortest_paths(str(root), str(l))[0]), leaves)
            )
        )
    )
    common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
    fork_t = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max()

    fork_id = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].idxmax()

    prefork_stat = list()
    for m in tqdm(
        range(n_map), disable=n_map == 1, file=sys.stdout, desc="    multi mapping "
    ):

        if n_map == 1:
            logg.info("    single mapping")
        df = adata.uns["pseudotime_list"][str(m)]
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        ddf = getpath(img, root, adata.uns["graph"]["tips"], fork_id, graph, df)

        ddf = ddf.loc[np.array(ddf.t.values) < (ddf.t.max() - pseudotime_offset)]

        df = adata.uns[name]["fork"]
        Xgenes = get_X(adata, ddf.index, df.index, layer, togenelist=True)
        data = list(zip([ddf] * len(Xgenes), Xgenes))

        allact = ProgressParallel(
            n_jobs=n_jobs,
            use_tqdm=n_map == 1,
            total=len(data),
            file=sys.stdout,
            desc="    prefork activation",
        )(delayed(test_prefork)(data[d]) for d in range(len(data)))

        allact = pd.DataFrame(
            np.array(allact), columns=["slope", "pval"], index=df.index
        )
        allact["fdr"] = multipletests(allact.pval, method="bonferroni")[1]
        allact["prefork_signi"] = (allact.fdr < fdr_cut) * 1

        prefork_stat = prefork_stat + [allact]

    df = adata.uns[name]["fork"]
    df_res = pd.DataFrame(
        {
            "slope": np.median([t.slope for t in prefork_stat], axis=0),
            "pval": np.median([t.pval for t in prefork_stat], axis=0),
            "fdr": np.median([t.fdr for t in prefork_stat], axis=0),
            "prefork_signi": np.mean([t.prefork_signi for t in prefork_stat], axis=0),
        },
        index=df.index,
    )
    df[["slope", "pre_pval", "pre_fdr", "prefork_signi"]] = df_res[
        ["slope", "pval", "fdr", "prefork_signi"]
    ]
    df["module"] = "late"
    df.loc[(df.slope > 0) & (df.prefork_signi > stf_cut), "module"] = "early"
    adata.uns[name]["fork"] = df

    for l in leaves:
        c_early = np.sum(
            (df.branch == str(keys[vals == l][0])) & (df.module == "early")
        )
        c_late = np.sum((df.branch == str(keys[vals == l][0])) & (df.module == "late"))
        logg.info(
            "    "
            + str(c_early)
            + " early and "
            + str(c_late)
            + " late features specific to leave "
            + str(keys[vals == l][0])
        )

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "updated \n"
        "    .uns['"
        + name
        + "']['fork'], DataFrame updated with additionnal 'slope','pval','fdr','prefork_signi' and 'module' columns."
    )

    return adata if copy else None


def test_prefork(dat):
    sdf = dat[0]
    sdf["exp"] = dat[1]
    result = sm.ols(formula="exp ~ t", data=sdf).fit()
    return [result.params["t"], result.pvalues["t"]]



from typing import Union, Optional
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from scanpy.plotting._utils import savefig_or_show
from scanpy import AnnData


def binned_pseudotime_meta(
    adata,
    key,
    nbins: int = 20,
    rotation: int = 0,
    show_colorbar: bool = False,
    rev=False,
    cmap="viridis",
    ax=None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
):

    """\
    Plot a dot plot of proportion of cells from a given category over binned sections of pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    key
        category to study.
    nbins
        Number of pseudotime bin to generate.
    rotation
        rotaton of the category labels.
    show_colorbar
        display pseudotime colorbar.
    cmap
        colormap of the pseudotime cbar.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a tuple of :class:`~matplotlib.axes.Axes`

    """

    ncats = len(adata.obs[key].cat.categories)
    intervals = pd.cut(adata.obs.t, bins=nbins)
    counts = pd.concat(
        [
            adata.obs[key][intervals[intervals == c].index].value_counts()
            for c in intervals.cat.categories
        ],
        axis=1,
    )
    prop = counts / counts.sum(axis=0)

    xs = np.arange(prop.shape[0])
    ys = np.arange(prop.shape[1])
    ys = [c.mid for c in intervals.cat.categories]
    ys = -np.array(ys) if rev else ys
    X, Y = np.meshgrid(xs, ys)

    specs = (
        {"width_ratios": [2.5 * ncats / 6, 0.25 * show_colorbar]}
        if show_colorbar
        else None
    )
    if ax is None:
        fig, ax = plt.subplots(
            1,
            1 + 1 * show_colorbar,
            figsize=(2.5 * ncats / 6 + 0.25 * show_colorbar, (nbins / 4) + 1),
            gridspec_kw=specs,
        )
        ax2 = ax[1] if show_colorbar else None
        ax = ax[0] if show_colorbar else ax

        if show_colorbar:
            cmap = eval("mpl.cm." + cmap)
            timebins = np.array([i.mid for i in intervals.cat.categories])
            norm = mpl.colors.Normalize(vmin=timebins.min(), vmax=timebins.max())
            norm = mpl.colors.BoundaryNorm(timebins, cmap.N)
            cbar = mpl.colorbar.ColorbarBase(
                ax2,
                cmap=cmap,
                norm=norm,
                spacing="proportional",
                orientation="vertical",
            )
            cbar.set_ticks([])

    if key + "_colors" not in adata.uns:
        from . import palette_tools

        palette_tools._set_default_colors_for_categorical_obs(adata, key)

    for i, d in enumerate(adata.obs[key].cat.categories):
        ax.scatter(
            X[:, i], Y[:, i], s=prop.values[i, :] * 200, c=adata.uns[key + "_colors"][i]
        )

    ax.grid(b=None)
    ax.set_xticks(range(len(adata.obs[key].cat.categories)))
    ax.set_xticklabels(adata.obs[key].cat.categories, rotation=rotation)
    ax.set_yticks([])
    ax.set_xlim([-0.5, ncats - 0.5])

    plt.tight_layout(pad=0.5)

    savefig_or_show("binned_pseudotime_meta", show=show, save=save)

    if show == False:
        return ax



import numpy as np
import pandas as pd
import igraph
from anndata import AnnData
import matplotlib.collections
from typing import Union, Optional, Sequence, Tuple, List
import plotly.graph_objects as go
import scanpy as sc
from cycler import Cycler

from pandas.api.types import is_categorical_dtype
from scanpy.plotting._utils import savefig_or_show

import matplotlib.patheffects as path_effects
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize, hex2color, rgb2hex
from numba import njit
import math

from .utils import is_categorical
from . import palette_tools
from ..tools.graph_operations import subset_tree
from .. import settings


def graph(
    adata: AnnData,
    basis: str = "umap",
    size_nodes: float = None,
    alpha_nodes: float = 1,
    linewidth: float = 2,
    alpha_seg: float = 1,
    color_cells: Union[str, None] = None,
    tips: bool = True,
    forks: bool = True,
    nodes: Optional[List] = [],
    ax=None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs,
):

    """\
    Project principal graph onto embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    basis
        Name of the `obsm` basis to use.
    size_nodes
        Size of the projected prinicpal points.
    alpha_nodes
        Alpha of nodes.
    linewidth
        Line width of the segments.
    alpha_seg
        Alpha of segments.
    color_cells
        cells color
    tips
        display tip ids.
    forks
        display fork ids.
    nodes
        display any node id.
    ax
        Add plot to existing ax
    show
        show the plot.
    save
        save the plot.
    kwargs
        arguments to pass to :func:`scanpy.pl.embedding`

    Returns
    -------
    If `show==False` a :class:`~matplotlib.axes.Axes`

    """

    if "graph" not in adata.uns:
        raise ValueError(
            "You need to run `tl.tree` or `tl.curve` first to compute a princal graph before plotting."
        )

    graph = adata.uns["graph"]

    emb = adata.obsm[f"X_{basis}"]

    if "components" in kwargs:
        cmp = np.array(kwargs["components"]) - 1
        emb = emb[:, cmp]

    else:
        emb = emb[:, :2]

    R = adata.obsm["X_R"]

    proj = (np.dot(emb.T, R) / R.sum(axis=0)).T

    B = graph["B"]

    if ax is None:
        ax = sc.pl.embedding(
            adata, color=color_cells, basis=basis, show=False, **kwargs
        )
    else:
        sc.pl.embedding(
            adata, color=color_cells, basis=basis, ax=ax, show=False, **kwargs
        )

    al = np.array(
        igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected").get_edgelist()
    )
    segs = al.tolist()
    vertices = proj.tolist()
    lines = [[tuple(vertices[j]) for j in i] for i in segs]
    lc = matplotlib.collections.LineCollection(
        lines, colors="k", linewidths=linewidth, alpha=alpha_seg
    )
    ax.add_collection(lc)

    ax.scatter(proj[:, 0], proj[:, 1], s=size_nodes, c="k", alpha=alpha_nodes)

    bbox = dict(facecolor="white", alpha=0.6, edgecolor="white", pad=0.1)

    if tips:
        for tip in graph["tips"]:
            ax.annotate(
                tip,
                (proj[tip, 0], proj[tip, 1]),
                ha="center",
                va="center",
                xytext=(-8, 8),
                textcoords="offset points",
                bbox=bbox,
            )
    if forks:
        for fork in graph["forks"]:
            ax.annotate(
                fork,
                (proj[fork, 0], proj[fork, 1]),
                ha="center",
                va="center",
                xytext=(-8, 8),
                textcoords="offset points",
                bbox=bbox,
            )
    if nodes:
        for node in nodes:
            ax.annotate(
                node,
                (proj[node, 0], proj[node, 1]),
                ha="center",
                va="center",
                xytext=(-8, 8),
                textcoords="offset points",
                bbox=bbox,
            )
    if show == False:
        return ax

    savefig_or_show("graph", show=show, save=save)


def trajectory(
    adata: AnnData,
    basis: str = "umap",
    root_milestone: Union[str, None] = None,
    milestones: Union[str, None] = None,
    color_seg: str = "t",
    cmap_seg: str = "viridis",
    layer_seg: Union[str, None] = "fitted",
    perc_seg: Union[List, None] = None,
    color_cells: Union[str, None] = None,
    scale_path: float = 1,
    arrows: bool = False,
    arrow_offset: int = 10,
    show_info: bool = True,
    ax=None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs,
):

    """\
    Project trajectory onto embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    basis
        Name of the `obsm` basis to use.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    col_seg
        color trajectory segments.
    layer_seg
        layer to use when coloring seg with a feature.
    perc_seg
        percentile cutoffs for segments.
    color_cells
        cells color.
    scale_path
        changes the width of the path
    arrows
        display arrows on segments (positioned at half pseudotime distance).
    arrow_offset
        arrow offset in number of nodes used to obtain its direction.
    show_info
        display legend/colorbar.
    ax
        Add plot to existing ax
    show
        show the plot.
    save
        save the plot.
    kwargs
        arguments to pass to :func:`scanpy.pl.embedding`.

    Returns
    -------
    If `show==False` a :class:`~matplotlib.axes.Axes`

    """

    if "graph" not in adata.uns:
        raise ValueError("You need to run `tl.pseudotime` first before plotting.")

    if (root_milestone is not None) & (color_seg == "milestones"):
        if ax is None:
            ax = sc.pl.embedding(adata, show=False, basis=basis, **kwargs)
        else:
            sc.pl.embedding(adata, show=False, ax=ax, basis=basis, **kwargs)
        verb_temp = settings.verbosity
        settings.verbosity = 1
        adata = subset_tree(
            adata,
            root_milestone="immature",
            milestones=["Ic", "II<>rest"],
            mode="extract",
            copy=True,
        )
        settings.verbosity = verb_temp

    graph = adata.uns["graph"]

    emb = adata.obsm[f"X_{basis}"]

    if "components" in kwargs:
        cmp = np.array(kwargs["components"]) - 1
        emb = emb[:, cmp]

    else:
        emb = emb[:, :2]

    R = adata.obsm["X_R"]

    nodes = graph["pp_info"].index
    proj = pd.DataFrame((np.dot(emb.T, R) / R.sum(axis=0)).T, index=nodes)

    B = graph["B"]
    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
    g.vs[:]["name"] = [v.index for v in g.vs]

    miles_ids = np.concatenate([graph["tips"], graph["forks"]])

    if (root_milestone is not None) & (color_seg != "milestones"):
        dct = graph["milestones"]
        nodes = g.get_all_shortest_paths(
            dct[root_milestone], [dct[m] for m in milestones]
        )
        nodes = np.unique(np.concatenate(nodes))
        tips = graph["tips"][np.isin(graph["tips"], nodes)]
        proj = proj.loc[nodes, :]
        g.delete_vertices(graph["pp_info"].index[~graph["pp_info"].index.isin(nodes)])
        if ax is None:
            ax = sc.pl.embedding(adata, show=False, basis=basis, **kwargs)
        else:
            sc.pl.embedding(adata, show=False, ax=ax, basis=basis, **kwargs)

    c_edges = np.array([e.split("|") for e in adata.obs.edge], dtype=int)
    cells = [any(np.isin(c_e, nodes)) for c_e in c_edges]

    import logging

    anndata_logger = logging.getLogger("anndata")
    prelog = anndata_logger.level
    anndata_logger.level = 40
    adata_c = adata[cells, :]

    if is_categorical(adata, color_cells):
        if (
            color_cells + "_colors" not in adata.uns
            or len(adata.uns[color_cells + "_colors"]) == 1
        ):

            palette_tools._set_default_colors_for_categorical_obs(adata, color_cells)

        adata_c.uns[color_cells + "_colors"] = [
            adata.uns[color_cells + "_colors"][
                np.argwhere(adata.obs[color_cells].cat.categories == c)[0][0]
            ]
            for c in adata_c.obs[color_cells].cat.categories
        ]

    if ax is None:
        ax = sc.pl.embedding(
            adata, color=color_cells, basis=basis, show=False, **kwargs
        )
    else:
        sc.pl.embedding(
            adata, color=color_cells, basis=basis, ax=ax, show=False, **kwargs
        )

    anndata_logger.level = prelog
    if show_info == False and color_cells is not None:
        remove_info(adata, ax, color_cells)

    al = np.array(g.get_edgelist())

    edges = [g.vs[e.tolist()]["name"] for e in al]

    lines = [[tuple(proj.loc[j]) for j in i] for i in edges]

    miles_ids = miles_ids[np.isin(miles_ids, proj.index)]

    if color_seg == "milestones":
        from matplotlib.colors import LinearSegmentedColormap

        rev_dict = dict(zip(graph["milestones"].values(), graph["milestones"].keys()))
        miles_cat = adata.obs.milestones.cat.categories
        mil_col = np.array(adata.uns["milestones_colors"])

        def get_milestones_gradients(i):
            start = graph["pp_seg"].iloc[i, 1]
            end = graph["pp_seg"].iloc[i, 2]
            mil_path = graph["pp_info"].time[g.get_all_shortest_paths(start, end)[0]]
            mil_path = (mil_path - mil_path.min()) / (mil_path.max() - mil_path.min())
            start_col = mil_col[miles_cat == rev_dict[start]][0]
            end_col = mil_col[miles_cat == rev_dict[end]][0]

            edges_mil = pd.Series(
                [
                    mil_path[[first, second]].mean()
                    for first, second in zip(mil_path.index, mil_path.index[1:])
                ],
                index=[
                    (first, second)
                    for first, second in zip(mil_path.index, mil_path.index[1:])
                ],
            )

            cmap = LinearSegmentedColormap.from_list("mil", [start_col, end_col])
            return pd.Series(
                [rgb2hex(c) for c in cmap(edges_mil)], index=edges_mil.index
            )

        edge_colors = pd.concat(
            [get_milestones_gradients(i) for i in range(graph["pp_seg"].shape[0])]
        )

        edges_tuples = [tuple(e) for e in edges]
        edge_colors.index = [
            e if any((np.array(edges_tuples) == e).sum(axis=1) == 2) else e[::-1]
            for e in edge_colors.index
        ]
        edge_colors = edge_colors[edges_tuples]

        color_segs = [hex2color(c) for c in edge_colors]
        color_mils = [
            mil_col[miles_cat == mm][0] for mm in [rev_dict[m] for m in miles_ids]
        ]

    else:
        vals = pd.Series(
            _get_color_values(adata, color_seg, layer=layer_seg)[0],
            index=adata.obs_names,
        )
        R = pd.DataFrame(adata.obsm["X_R"], index=adata.obs_names)
        R = R.loc[adata.obs_names]
        vals = vals[~np.isnan(vals)]
        R = R.loc[vals.index]

        def get_nval(i):
            return np.average(vals, weights=R.loc[:, i])

        node_vals = np.array(list(map(get_nval, range(R.shape[1]))))
        seg_val = node_vals[np.array(edges)].mean(axis=1)

        if perc_seg is not None:
            min_v, max_v = np.percentile(seg_val, perc_seg)
            seg_val[seg_val < min_v] = min_v
            seg_val[seg_val > max_v] = max_v

        sm = ScalarMappable(
            norm=Normalize(vmin=seg_val.min(), vmax=seg_val.max()), cmap=cmap_seg
        )

        lines = [lines[i] for i in np.argsort(seg_val)]
        seg_val = seg_val[np.argsort(seg_val)]

        color_segs = [sm.to_rgba(sv) for sv in seg_val]
        color_mils = sm.to_rgba(node_vals[miles_ids])

    lc = matplotlib.collections.LineCollection(
        lines,
        colors="k",
        linewidths=7.5 * scale_path,
        zorder=100,
        path_effects=[path_effects.Stroke(capstyle="round")],
    )
    ax.add_collection(lc)

    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
    seg = graph["pp_seg"].loc[:, ["from", "to"]].values.tolist()

    if arrows:
        for s in seg:
            path = np.array(g.get_shortest_paths(s[0], s[1])[0])
            coord = proj.loc[
                path,
            ].values
            out = np.empty(len(path) - 1)
            cdist_numba(coord, out)
            mid = np.argmin(np.abs(out.cumsum() - out.sum() / 2))
            if mid + arrow_offset > (len(path) - 1):
                arrow_offset = len(path) - 1 - mid
            ax.quiver(
                proj.loc[path[mid], 0],
                proj.loc[path[mid], 1],
                proj.loc[path[mid + arrow_offset], 0] - proj.loc[path[mid], 0],
                proj.loc[path[mid + arrow_offset], 1] - proj.loc[path[mid], 1],
                headwidth=15 * scale_path,
                headaxislength=10 * scale_path,
                headlength=10 * scale_path,
                units="dots",
                zorder=101,
            )
            c_arrow = vals[
                (
                    np.array([e.split("|") for e in adata.obs.edge], dtype=int)
                    == path[mid]
                ).sum(axis=1)
                == 1
            ].mean()
            ax.quiver(
                proj.loc[path[mid], 0],
                proj.loc[path[mid], 1],
                proj.loc[path[mid + arrow_offset], 0] - proj.loc[path[mid], 0],
                proj.loc[path[mid + arrow_offset], 1] - proj.loc[path[mid], 1],
                headwidth=12 * scale_path,
                headaxislength=10 * scale_path,
                headlength=10 * scale_path,
                units="dots",
                color=sm.to_rgba(c_arrow),
                zorder=102,
            )

    lc = matplotlib.collections.LineCollection(
        lines,
        colors=color_segs,
        linewidths=5 * scale_path,
        zorder=104,
        path_effects=[path_effects.Stroke(capstyle="round")],
    )

    ax.scatter(
        proj.loc[miles_ids, 0],
        proj.loc[miles_ids, 1],
        zorder=103,
        c="k",
        s=200 * scale_path,
    )
    ax.add_collection(lc)

    ax.scatter(
        proj.loc[miles_ids, 0],
        proj.loc[miles_ids, 1],
        zorder=105,
        c=color_mils,
        s=140 * scale_path,
    )

    if show == False:
        return ax

    savefig_or_show("trajectory", show=show, save=save)


@njit()
def cdist_numba(coords, out):
    for i in range(0, coords.shape[0] - 1):
        out[i] = math.sqrt(
            (coords[i, 0] - coords[i + 1, 0]) ** 2
            + (coords[i, 1] - coords[i + 1, 1]) ** 2
        )


def scatter3d(emb, col, cell_size, nm):
    return go.Scatter3d(
        x=emb[:, 0],
        y=emb[:, 1],
        z=emb[:, 2],
        mode="markers",
        marker=dict(size=cell_size, color=col, opacity=0.9),
        name=nm,
    )


def trajectory_3d(
    adata: AnnData,
    basis: str = "umap3d",
    color: str = None,
    traj_width: int = 5,
    cell_size: int = 2,
    figsize: tuple = (900, 900),
    cmap=None,
):

    """\
    Project trajectory onto 3d embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    basis
        Name of the `obsm` basis to use.
    color
        cells color.
    traj_width
        segments width.
    cell_size
        cell size.
    figsize
        figure size in pixels.
    cmap
        colormap of the cells.

    Returns
    -------
    an interactive plotly graph figure.

    """

    r = adata.uns["graph"]

    emb = adata.obsm[f"X_{basis}"]
    if emb.shape[1] > 3:
        raise ValueError("Embedding is not three dimensional.")

    R = adata.obsm["X_R"]
    proj = (np.dot(emb.T, R) / R.sum(axis=0)).T

    B = r["B"]

    al = np.array(
        igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected").get_edgelist()
    )
    segs = al.tolist()
    vertices = proj.tolist()
    vertices = np.array(vertices)
    segs = np.array(segs)

    x_lines = list()
    y_lines = list()
    z_lines = list()

    x = vertices[:, 0]
    y = vertices[:, 1]
    z = vertices[:, 2]

    for i in range(segs.shape[0]):
        p = segs[i, :]
        for i in range(2):
            x_lines.append(x[p[i]])
            y_lines.append(y[p[i]])
            z_lines.append(z[p[i]])
        x_lines.append(None)
        y_lines.append(None)
        z_lines.append(None)

    if color is not None:
        if adata.obs[color].dtype.name == "str":
            adata.obs[color] = adata.obs[color].astype("category")

        if color + "_colors" not in adata.uns:
            palette_tools._set_default_colors_for_categorical_obs(adata, color)

        if adata.obs[color].dtype.name == "category":
            pal_dict = dict(
                zip(adata.obs[color].cat.categories, adata.uns[color + "_colors"])
            )
            trace1 = list(
                map(
                    lambda x: scatter3d(
                        emb[adata.obs[color] == x, :], pal_dict[x], cell_size, x
                    ),
                    list(pal_dict.keys()),
                )
            )

        else:
            if cmap is None:
                cmap = "Viridis"
            trace1 = [
                go.Scatter3d(
                    x=emb[:, 0],
                    y=emb[:, 1],
                    z=emb[:, 2],
                    mode="markers",
                    marker=dict(
                        size=cell_size,
                        color=adata.obs[color],
                        colorscale=cmap,
                        opacity=0.9,
                    ),
                )
            ]

    else:
        trace1 = [
            go.Scatter3d(
                x=emb[:, 0],
                y=emb[:, 1],
                z=emb[:, 2],
                mode="markers",
                marker=dict(size=cell_size, color="grey", opacity=0.9),
            )
        ]

    trace2 = [
        go.Scatter3d(
            x=x_lines,
            y=y_lines,
            z=z_lines,
            mode="lines",
            name="lines",
            line=dict(width=traj_width),
        )
    ]

    fig = go.Figure(data=trace1 + trace2)
    fig.update_layout(
        scene=dict(
            xaxis=dict(visible=False),
            yaxis=dict(visible=False),
            zaxis=dict(visible=False),
        ),
        width=figsize[0],
        height=figsize[0],
        margin=dict(l=5, r=5, t=5, b=5),
    )
    fig.show()


def _get_color_values(
    adata,
    value_to_plot,
    groups=None,
    palette: Union[str, Sequence[str], Cycler, None] = None,
    use_raw=False,
    gene_symbols=None,
    layer=None,
) -> Tuple[Union[np.ndarray, str], bool]:
    """
    Returns the value or color associated to each data point.
    For categorical data, the return value is list of colors taken
    from the category palette or from the given `palette` value.
    For non-categorical data, the values are returned
    Returns
    -------
    values
        Values to plot
    is_categorical
        Are the values categorical?
    """
    if value_to_plot is None:
        return "lightgray", False
    if (
        gene_symbols is not None
        and value_to_plot not in adata.obs.columns
        and value_to_plot not in adata.var_names
    ):
        # We should probably just make an index for this, and share it over runs
        value_to_plot = adata.var.index[adata.var[gene_symbols] == value_to_plot][
            0
        ]  # TODO: Throw helpful error if this doesn't work
    if use_raw and value_to_plot not in adata.obs.columns:
        values = adata.raw.obs_vector(value_to_plot)
    else:
        values = adata.obs_vector(value_to_plot, layer=layer)

    ###
    # when plotting, the color of the dots is determined for each plot
    # the data is either categorical or continuous and the data could be in
    # 'obs' or in 'var'
    if not is_categorical_dtype(values):
        return values, False
    else:  # is_categorical_dtype(values)
        color_key = f"{value_to_plot}_colors"
        if palette:
            palette_tools._set_colors_for_categorical_obs(adata, value_to_plot, palette)
        elif color_key not in adata.uns or len(adata.uns[color_key]) < len(
            values.categories
        ):
            #  set a default palette in case that no colors or few colors are found
            palette_tools._set_default_colors_for_categorical_obs(adata, value_to_plot)
        else:
            palette_tools._validate_palette(adata, value_to_plot)

        color_vector = np.asarray(adata.uns[color_key])[values.codes]

        # Handle groups
        if groups:
            color_vector = np.fromiter(
                map(colors.to_hex, color_vector), "<U15", len(color_vector)
            )
            # set color to 'light gray' for all values
            # that are not in the groups
            color_vector[~adata.obs[value_to_plot].isin(groups)] = "lightgray"
        return color_vector, True


def remove_info(adata, ax, color):
    if is_categorical(adata, color):
        if ax.get_legend() is not None:
            ax.get_legend().remove()
    else:
        ax.set_box_aspect(aspect=1)
        fig = ax.get_gridspec().figure
        cbar = np.argwhere(
            ["colorbar" in a.get_label() for a in fig.get_axes()]
        ).ravel()
        if len(cbar) > 0:
            fig.get_axes()[cbar[0]].remove()



from .graph_operations import (
    cleanup,
    getpath,
    subset_tree,
    attach_tree,
    extend_tips,
    simplify,
)
from .graph_fitting import tree, curve, circle
from .root import root, roots
from .pseudotime import pseudotime, rename_milestones
from .test_association import test_association
from .fit import fit
from .cluster import cluster
from .slide_cors import slide_cells, slide_cors
from .correlation_tools import (
    synchro_path,
    synchro_path_multi,
    module_inclusion,
)
from .bifurcation_tools import test_fork, branch_specific, activation, activation_lm

# from .activation_tools import activation, activation_lm
from .conversion import cellrank_to_tree
from .linearity_deviation import linearity_deviation
from .dendrogram import dendrogram



from typing import Optional, Union
from anndata import AnnData
import numpy as np
import pandas as pd
import igraph

import itertools
from sklearn.metrics import pairwise_distances
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path

from .utils import get_X

from .. import logging as logg
from .. import settings


def root(
    adata: AnnData,
    root: Union[int, str],
    tips_only: bool = False,
    min_val: bool = False,
    layer: Optional = None,
    copy: bool = False,
):
    """\
    Define the root of the trajectory.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root
        Either an Id (int) of the tip of the fork to be considered as a root. Or a key (str) from obs/X (such as CytoTRACE) for automatic selection.
    tips_only
        Perform automatic assignment on tips only.
    min_val
        Perform automatic assignment using minimum value instead.
    layer
        If key is in X, choose which layer to use for the averaging.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['graph']['root']`
            selected root.
        `.uns['graph']['pp_info']`
            for each PP, its distance vs root and segment assignment.
        `.uns['graph']['pp_seg']`
            segments network information.
    """

    adata = adata.copy() if copy else adata

    if "graph" not in adata.uns:
        raise ValueError(
            "You need to run `tl.tree` or `tl.curve` first to compute a princal graph before choosing a root."
        )

    graph = adata.uns["graph"]
    circle = False
    if type(root) == str:
        if root in adata.obs:
            root_val = adata.obs[root]
        if root in adata.var_names:
            root_val = get_X(adata, adata.obs_names, root, layer).ravel()

        logg.info("automatic root selection using " + root + " values", time=False)
        avgs = list(
            map(
                lambda n: np.average(root_val, weights=adata.obsm["X_R"][:, n]),
                range(adata.obsm["X_R"].shape[1]),
            )
        )
        avgs = np.array(avgs)
        if tips_only:
            mask = np.ones(avgs.shape, bool)
            mask[adata.uns["graph"]["tips"]] = False
            avgs[mask] = 0
        if min_val:
            if tips_only:
                avgs[mask] = avgs.max()
            root = np.argmin(avgs)
        else:
            root = np.argmax(avgs)

        if len(graph["tips"]) == 0:
            B = adata.uns["graph"]["B"]
            g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
            circle = True
            if min_val:
                todel = g.neighbors(root)[
                    np.argmax([avgs[i] for i in g.neighbors(root)])
                ]
            else:
                todel = g.neighbors(root)[
                    np.argmin([avgs[i] for i in g.neighbors(root)])
                ]
            B[root, todel] = 0
            B[todel, root] = 0
            g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
            # Add edge weights and node labels.
            g.es["weight"] = B[B.nonzero()]
            tips = np.argwhere(np.array(g.degree()) == 1).flatten()
            forks = np.argwhere(np.array(g.degree()) > 2).flatten()
            adata.uns["graph"]["B"] = B
            adata.uns["graph"]["tips"] = tips

    d = 1e-6 + pairwise_distances(graph["F"].T, graph["F"].T, metric=graph["metrics"])

    to_g = graph["B"] * d

    csr = csr_matrix(to_g)

    g = igraph.Graph.Adjacency((to_g > 0).tolist(), mode="undirected")
    g.es["weight"] = to_g[to_g.nonzero()]

    root_dist_matrix = shortest_path(csr, directed=False, indices=root)
    pp_info = pd.DataFrame(
        {"PP": g.vs.indices, "time": root_dist_matrix, "seg": np.zeros(csr.shape[0])}
    )

    nodes = np.argwhere(
        np.apply_along_axis(arr=(csr > 0).todense(), axis=0, func1d=np.sum) != 2
    ).flatten()
    nodes = np.unique(np.append(nodes, root))

    pp_seg = pd.DataFrame(columns=["n", "from", "to", "d"])
    for node1, node2 in itertools.combinations(nodes, 2):
        paths12 = g.get_shortest_paths(node1, node2)
        paths12 = np.array([val for sublist in paths12 for val in sublist])

        if np.sum(np.isin(nodes, paths12)) == 2:
            fromto = np.array([node1, node2])
            path_root = root_dist_matrix[[node1, node2]]
            fro = fromto[np.argmin(path_root)]
            to = fromto[np.argmax(path_root)]
            pp_info.loc[paths12, "seg"] = pp_seg.shape[0] + 1
            pp_seg = pp_seg.append(
                pd.DataFrame(
                    {
                        "n": pp_seg.shape[0] + 1,
                        "from": fro,
                        "to": to,
                        "d": shortest_path(csr, directed=False, indices=fro)[to],
                    },
                    index=[pp_seg.shape[0] + 1],
                )
            )

    pp_seg["n"] = pp_seg["n"].astype(int).astype(str)
    pp_seg["n"] = pp_seg["n"].astype(int).astype(str)

    pp_seg["from"] = pp_seg["from"].astype(int)
    pp_seg["to"] = pp_seg["to"].astype(int)

    pp_info["seg"] = pp_info["seg"].astype(int).astype(str)
    pp_info["seg"] = pp_info["seg"].astype(int).astype(str)

    graph["pp_info"] = pp_info
    graph["pp_seg"] = pp_seg
    graph["root"] = root

    adata.uns["graph"] = graph

    logg.info(
        "node " + str(root) + " selected as a root",
        time=False,
        end=" " if settings.verbosity > 2 else "\n",
    )
    logg.hint(
        "added\n"
        "    .uns['graph']['root'] selected root.\n"
        "    .uns['graph']['pp_info'] for each PP, its distance vs root and segment assignment.\n"
        "    .uns['graph']['pp_seg'] segments network information."
    )

    if circle:
        logg.hint(
            "updated\n"
            "    .uns['graph']['B'] with circle now converted to curved trajectory.\n"
            "    .uns['graph']['tips'] new tips delimitating the trajectory."
        )

    return adata if copy else None


def roots(adata: AnnData, roots, meeting, copy: bool = False):

    """\
    Define 2 roots of the tree.

    Parameters
    ----------
    adata
        Annotated data matrix.
    roots
        list of tips or forks to be considered a roots.
    meeting
         node ID of the meeting point fo the two converging paths.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.uns['graph']['root']`
            farthest root selected.
        `.uns['graph']['root2']`
            2nd root selected.
        `.uns['graph']['meeting']`
            meeting point on the tree.
        `.uns['graph']['pp_info']`
            for each PP, its distance vs root and segment assignment).
        `.uns['graph']['pp_seg']`
            segments network information.
    """

    adata = adata.copy() if copy else adata

    if "graph" not in adata.uns:
        raise ValueError(
            "You need to run `tl.tree` first to compute a princal tree before choosing two roots."
        )

    graph = adata.uns["graph"]

    d = 1e-6 + pairwise_distances(graph["F"].T, graph["F"].T, metric=graph["metrics"])

    to_g = graph["B"] * d

    csr = csr_matrix(to_g)

    g = igraph.Graph.Adjacency((to_g > 0).tolist(), mode="undirected")
    g.es["weight"] = to_g[to_g.nonzero()]

    root = roots[
        np.argmax(shortest_path(csr, directed=False, indices=roots)[:, meeting])
    ]
    root2 = roots[
        np.argmin(shortest_path(csr, directed=False, indices=roots)[:, meeting])
    ]

    root_dist_matrix = shortest_path(csr, directed=False, indices=root)
    pp_info = pd.DataFrame(
        {"PP": g.vs.indices, "time": root_dist_matrix, "seg": np.zeros(csr.shape[0])}
    )

    nodes = np.argwhere(
        np.apply_along_axis(arr=(csr > 0).todense(), axis=0, func1d=np.sum) != 2
    ).flatten()
    pp_seg = pd.DataFrame(columns=["n", "from", "to", "d"])
    for node1, node2 in itertools.combinations(nodes, 2):
        paths12 = g.get_shortest_paths(node1, node2)
        paths12 = np.array([val for sublist in paths12 for val in sublist])

        if np.sum(np.isin(nodes, paths12)) == 2:
            fromto = np.array([node1, node2])
            path_root = root_dist_matrix[[node1, node2]]
            fro = fromto[np.argmin(path_root)]
            to = fromto[np.argmax(path_root)]
            pp_info.loc[paths12, "seg"] = pp_seg.shape[0] + 1
            pp_seg = pp_seg.append(
                pd.DataFrame(
                    {
                        "n": pp_seg.shape[0] + 1,
                        "from": fro,
                        "to": to,
                        "d": shortest_path(csr, directed=False, indices=fro)[to],
                    },
                    index=[pp_seg.shape[0] + 1],
                )
            )

    pp_seg["n"] = pp_seg["n"].astype(int).astype(str)
    pp_seg["n"] = pp_seg["n"].astype(int).astype(str)

    pp_info["seg"] = pp_info["seg"].astype(int).astype(str)
    pp_info["seg"] = pp_info["seg"].astype(int).astype(str)

    tips = graph["tips"]
    tips = tips[~np.isin(tips, roots)]

    edges = pp_seg[["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(np.unique(pp_seg[["from", "to"]].values.flatten().astype(str)))
    img.add_edges(edges)

    root2paths = pd.Series(
        shortest_path(csr, directed=False, indices=root2)[tips.tolist() + [meeting]],
        index=tips.tolist() + [meeting],
    )

    toinvert = root2paths.index[(root2paths <= root2paths[meeting])]

    for toinv in toinvert:
        pathtorev = np.array(img.vs[:]["name"])[
            np.array(img.get_shortest_paths(str(root2), str(toinv)))
        ][0]
        for i in range((len(pathtorev) - 1)):
            segtorev = pp_seg.index[
                pp_seg[["from", "to"]]
                .astype(str)
                .apply(lambda x: all(x.values == pathtorev[[i + 1, i]]), axis=1)
            ]

            pp_seg.loc[segtorev, ["from", "to"]] = pp_seg.loc[segtorev][
                ["to", "from"]
            ].values
            pp_seg["from"] = pp_seg["from"].astype(int)
            pp_seg["to"] = pp_seg["to"].astype(int)

    pptoinvert = np.unique(np.concatenate(g.get_shortest_paths(root2, toinvert)))
    reverted_dist = (
        shortest_path(csr, directed=False, indices=root2)
        + np.abs(
            np.diff(shortest_path(csr, directed=False, indices=roots)[:, meeting])
        )[0]
    )
    pp_info.loc[pptoinvert, "time"] = reverted_dist[pptoinvert]

    graph["pp_info"] = pp_info
    graph["pp_seg"] = pp_seg
    graph["root"] = root
    graph["root2"] = root2
    graph["meeting"] = meeting

    adata.uns["graph"] = graph

    logg.info("root selected", time=False, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added\n" + "    " + str(root) + " is the farthest root.\n"
        "    .uns['graph']['root'] farthest root selected.\n"
        "    .uns['graph']['root2'] 2nd root selected.\n"
        "    .uns['graph']['meeting'] meeting point on the tree.\n"
        "    .uns['graph']['pp_info'] for each PP, its distance vs root and segment assignment.\n"
        "    .uns['graph']['pp_seg'] segments network information."
    )

    return adata if copy else None



from anndata import AnnData
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex
from typing import Union, Optional
import numpy as np
from scanpy.plotting._utils import savefig_or_show


def test_fork(
    adata: AnnData,
    root_milestone,
    milestones,
    col: Union[None, list] = None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
):
    """\
    Plot results generated from tl.test_fork.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    col
        color of the two sets of genes, by default to the color of the end milestones.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    if "milestones_colors" not in adata.uns or len(adata.uns["milestones_colors"]) == 1:
        from . import palette_tools

        palette_tools._set_default_colors_for_categorical_obs(adata, "milestones")

    mlsc = np.array(adata.uns["milestones_colors"].copy())
    if mlsc.dtype == "float":
        mlsc = list(map(rgb2hex, mlsc))

    name = root_milestone + "->" + "<>".join(milestones)
    df = adata.uns[name]["fork"]
    df = df.loc[df.fdr < 0.05]

    c_mil = (
        np.array(mlsc)[
            np.argwhere(adata.obs.milestones.cat.categories.isin(milestones))
        ].flatten()
        if col is None
        else col
    )

    A = df.iloc[:, [0, 1]].values

    fig, ax = plt.subplots()

    ax.scatter(A[A[:, 0] != 0, 0], df.loc[A[:, 0] != 0].fdr.values, color=c_mil[0])
    ax.scatter(
        np.abs(A[A[:, 1] != 0, 1]), df.loc[A[:, 1] != 0].fdr.values, color=c_mil[1]
    )

    ax.set_xlabel(" < A > ".join(milestones))
    ax.set_ylabel("FDR")

    left, right = ax.get_xlim()
    bounds = np.abs(A.ravel()).max()
    bounds = bounds + bounds * 0.1
    ax.set_xlim([-bounds, bounds])
    xticks = ax.get_xticks()
    xticks = [str(int(xt)) if xt.is_integer() else str(xt) for xt in np.abs(xticks)]
    ax.set_xticklabels(xticks)

    if show == False:
        return ax

    savefig_or_show("test_fork", show=show, save=save)



from .trajectory import graph, trajectory, trajectory_3d
from .features import trends, single_trend
from .test_association import test_association
from .modules import modules
from .milestones import milestones
from .slide_cors import slide_cors
from .synchro_path import synchro_path
from .test_fork import test_fork
from .module_inclusion import module_inclusion
from .linearity_deviation import linearity_deviation
from .dendrogram import dendrogram
from .binned_pseudotime_meta import binned_pseudotime_meta
from .matrix import matrix



from typing_extensions import Literal
import igraph
import numpy as np
import pandas as pd
from typing import Union
import matplotlib
import matplotlib.cm as cm
import matplotlib.colors as mcolors

from ..tools.dendrogram import hierarchy_pos
from .. import logging as logg


def milestones(
    adata,
    color=None,
    cmap=None,
    roots=None,
    layout: Literal["dendro", "reingold_tilford"] = "dendro",
    figsize=(500, 500),
):
    """\
    Display the milestone graph in PAGA style.

    Parameters
    ----------
    adata
        Annotated data matrix.
    color
        color the milestones with variable from adata.obs.
    cmap
        colormap to use for the node coloring.
    roots
        select milestones to position on top fo the plot.
    dendro
        generate layout following dendrogram representation.
    figsize
        figure size in pixels

    Returns
    -------
    igraph.plot

    """

    graph = adata.uns["graph"]

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph(directed=True)
    img.add_vertices(vals.astype(str))
    img.add_edges(edges)

    img.vs["label"] = keys

    dct = dict(zip(img.vs["name"], img.vs["label"]))
    if roots is None:
        if "root2" not in adata.uns["graph"]:
            roots = [dct[str(adata.uns["graph"]["root"])]]
        else:
            roots = [
                dct[str(adata.uns["graph"]["root"])],
                dct[str(adata.uns["graph"]["root2"])],
            ]
            if layout == "dendro":
                logg.warn("two roots detected, reverting to reingold_tilford layout")
                layout = "reingold_tilford"

    if color is None:
        if "milestones_colors" not in adata.uns:
            from . import palette_tools

            palette_tools._set_default_colors_for_categorical_obs(adata, "milestones")
        img.vs["color"] = [
            np.array(adata.uns["milestones_colors"])[
                adata.obs.milestones.cat.categories == dct[m]
            ][0]
            for m in img.vs["name"]
        ]
    else:
        if cmap is None:
            cmap = "viridis"
        g = adata.obs.groupby("milestones")
        val_milestones = g.apply(lambda x: np.mean(x[color]))
        norm = matplotlib.colors.Normalize(
            vmin=min(val_milestones), vmax=max(val_milestones), clip=True
        )
        mapper = cm.ScalarMappable(norm=norm, cmap=eval("cm." + cmap))
        c_mil = list(
            map(lambda m: mcolors.to_hex(mapper.to_rgba(m)), val_milestones.values)
        )
        img.vs["color"] = c_mil

    if layout == "dendro":
        pos = hierarchy_pos(img.to_networkx(), vert_gap=0.1)
        pos = pd.DataFrame([pos[s] for s in np.sort(np.array(list(pos.keys())))])
        pos.iloc[:, 1] = -pos.iloc[:, 1]
        layout = list(pos.to_records(index=False))
    else:
        layout = img.layout_reingold_tilford(
            root=list(
                map(
                    lambda root: np.argwhere(np.array(img.vs["label"]) == root)[0][0],
                    roots,
                )
            )
        )

    return igraph.plot(img, bbox=figsize, layout=layout, margin=50)



import numpy as np
import pandas as pd
from anndata import AnnData

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.colors import rgb2hex
from matplotlib.patches import ConnectionPatch

from typing import Union, Optional
from scanpy.plotting._utils import savefig_or_show
from scFates.plot.dendrogram import dendrogram
from adjustText import adjust_text
from matplotlib import gridspec
from sklearn.metrics import pairwise_distances

from .. import get


def slide_cors(
    adata: AnnData,
    root_milestone,
    milestones,
    col: Union[None, list] = None,
    basis: str = "umap",
    win_keep: Union[None, list] = None,
    frame_emb: bool = True,
    focus=None,
    top_focus=4,
    labels: Union[None, tuple] = None,
    fig_height: float = 6,
    fontsize: int = 16,
    fontsize_focus: int = 18,
    point_size: int = 20,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    kwargs_text: dict = {},
    kwargs_con: dict = {},
    kwargs_adjust: dict = {},
    **kwargs,
):

    """\
    Plot results generated from tl.slide_cors.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    genesetA
        plot correlation with custom geneset.
    genesetB
        plot correlation with custom geneset.
    col
        specify color for the two modules, by default according to their respective milestones.
    basis
        Name of the `obsm` basis to use.
    win_keep
        plot only a subset of windows.
    frame_emb
        add frame around emb plot.
    focus
        add on the right side a scatter focusing one defined window.
    top_focus
        highlight n top markers for each module, having the greatest distance to 0,0 coordinates.
    labels
        labels defining the two modules, named after the milestones if None, or 'A' and 'B' if less than two milestones is used.
    fig_height
        figure height.
    fontsize
        repulsion score font size.
    fontsize_focus
        fontsize of x and y labels in focus plot.
    point_size
        correlation plot point size.
    show
        show the plot.
    save
        save the plot.
    kwargs_text
        parameters for the text annotation of the labels.
    kwargs_con
        parameters passed on the ConnectionPatch linking the focused plot to the rest.
    kwargs_adjust
        parameters passed to adjust_text.
    **kwargs
        if `basis=dendro`, arguments passed to :func:`scFates.pl.dendrogram`

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    if "milestones_colors" not in adata.uns or len(adata.uns["milestones_colors"]) == 1:
        from . import palette_tools

        palette_tools._set_default_colors_for_categorical_obs(adata, "milestones")

    mlsc = np.array(adata.uns["milestones_colors"].copy())
    if mlsc.dtype == "float":
        mlsc = list(map(rgb2hex, mlsc))

    name = root_milestone + "->" + "<>".join(milestones)
    freqs = adata.uns[name]["cell_freq"]
    nwin = len(freqs)

    corAB = adata.uns[name]["corAB"].copy()

    if len(milestones) > 1:
        genesetA = corAB[milestones[0]]["genesetA"].index
        genesetB = corAB[milestones[0]]["genesetB"].index
        corA = pd.concat(adata.uns[name]["corAB"][milestones[0]])
        corB = pd.concat(adata.uns[name]["corAB"][milestones[1]])
        if labels is None:
            labelA, labelB = milestones if labels is None else labels
    else:
        genesetA = corAB["A"]["genesetA"].index
        genesetB = corAB["A"]["genesetB"].index
        corA = pd.concat(adata.uns[name]["corAB"]["A"])
        corB = pd.concat(adata.uns[name]["corAB"]["B"])
        labelA, labelB = ("A", "B") if labels is None else labels

    groupsA = np.ones(corA.shape[0])
    groupsA[: len(genesetA)] = 2
    groupsB = np.ones(corA.shape[0])
    groupsB[len(genesetA) :] = 2

    gr = LinearSegmentedColormap.from_list("greyreds", ["lightgrey", "black"])

    if win_keep is not None:
        freqs = [freqs[i] for i in win_keep]
        corA = corA.iloc[:, win_keep]
        corB = corB.iloc[:, win_keep]
        corA.columns = np.arange(len(win_keep)).astype(str)
        corB.columns = np.arange(len(win_keep)).astype(str)

    maxlim = (
        np.max(
            [
                corB.max().max(),
                np.abs(corB.min().min()),
                corA.max().max(),
                np.abs(corA.min().min()),
            ]
        )
        + 0.01
    )

    nwin = len(freqs)

    focus_true = False if focus is None else True

    fig = plt.figure(figsize=((nwin + focus_true * 2) * fig_height / 2, fig_height))

    gs0 = gridspec.GridSpec(
        1,
        1 + focus_true * 1,
        figure=fig,
        width_ratios=[nwin, 2] if focus_true else None,
        wspace=0.05,
    )

    gs00 = gridspec.GridSpecFromSubplotSpec(
        2, nwin, subplot_spec=gs0[0], hspace=0.05, wspace=0.05
    )

    emb = adata.obsm["X_" + basis]

    for i in range(nwin):
        ax_emb = fig.add_subplot(gs00[0, i])
        freq = freqs[i]
        if basis == "dendro":
            if "s" in kwargs:
                s = kwargs["s"]
                del kwargs["s"]
            else:
                s = 10
            dendrogram(
                adata,
                s=s,
                cmap=gr,
                show=False,
                ax=ax_emb,
                show_info=False,
                title="",
                alpha=0,
                **kwargs,
            )
            ax_emb.set_xlabel("")
            ax_emb.set_ylabel("")

        ax_emb.scatter(
            emb[np.argsort(freq), 0],
            emb[np.argsort(freq), 1],
            s=10,
            c=freq[np.argsort(freq)],
            cmap=gr,
            rasterized=True,
        )
        ax_emb.grid(b=None)
        ax_emb.set_xticks([])
        ax_emb.set_yticks([])
        ma = np.max(plt.rcParams["figure.figsize"])
        mi = np.min(plt.rcParams["figure.figsize"])
        toreduce = plt.rcParams["figure.figsize"] / ma
        if len(np.argwhere(toreduce != 1)) == 0:
            pass
        elif np.argwhere(toreduce != 1)[0][0] == 0:
            span = ax_emb.get_xlim()[1] - ax_emb.get_xlim()[0]
            midpoint = ax_emb.get_xlim()[0] + span / 2
            enlargment = (ax_emb.get_xlim()[1] - ax_emb.get_xlim()[0]) * (ma / mi) / 2
            ax_emb.set_xlim([midpoint - enlargment, midpoint + enlargment])
        elif np.argwhere(toreduce != 1)[0][0] == 1:
            span = ax_emb.get_ylim()[1] - ax_emb.get_ylim()[0]
            midpoint = ax_emb.get_ylim()[0] + span / 2
            enlargment = (ax_emb.get_ylim()[1] - ax_emb.get_ylim()[0]) * (ma / mi) / 2
            ax_emb.set_ylim([midpoint - enlargment, midpoint + enlargment])
        if frame_emb == False:
            ax_emb.axis("off")

    c_mil = (
        [
            np.array(mlsc)[adata.obs.milestones.cat.categories == miles]
            for miles in milestones
        ]
        if len(milestones) == 2
        else ["tab:blue", "tab:red"]
    )
    c_mil = col if col is not None else c_mil
    genesets = [genesetA, genesetB]
    for i in range(nwin):
        ax_scat = fig.add_subplot(gs00[1, i])
        for j in range(2):
            ax_scat.scatter(
                corA.loc[groupsB == (j + 1), str(i)],
                corB.loc[groupsB == (j + 1), str(i)],
                color=c_mil[j],
                alpha=0.5,
                rasterized=True,
                s=point_size,
            )
        rep = (
            np.corrcoef(groupsA, corA.iloc[:, i])[0][1]
            + np.corrcoef(groupsB, corB.iloc[:, i])[0][1]
        ) / 2
        ax_scat.annotate(
            str(round(rep, 2)),
            xy=(0.7, 0.88),
            xycoords="axes fraction",
            fontsize=fontsize,
        )
        ax_scat.grid(b=None)
        ax_scat.axvline(0, linestyle="dashed", color="grey", zorder=0)
        ax_scat.axhline(0, linestyle="dashed", color="grey", zorder=0)
        ax_scat.set_xlim([-maxlim, maxlim])
        ax_scat.set_ylim([-maxlim, maxlim])
        ax_scat.set_xticks([])
        ax_scat.set_yticks([])
        if i == focus:
            ax_focus = ax_scat
            ax_focus.tick_params(color="red", labelcolor="red")
            for spine in ax_focus.spines.values():
                spine.set_edgecolor("red")
                spine.set_linewidth(spine.get_linewidth() * 2)

    if focus_true:
        ax_scat = fig.add_subplot(gs0[1])
        for j in range(2):
            ax_scat.scatter(
                corA.loc[groupsB == (j + 1), str(focus)],
                corB.loc[groupsB == (j + 1), str(focus)],
                color=c_mil[j],
                alpha=0.5,
                rasterized=True,
                s=point_size * 2,
            )
        ax_scat.grid(b=None)
        ax_scat.axvline(0, linestyle="dashed", color="grey", zorder=0)
        ax_scat.axhline(0, linestyle="dashed", color="grey", zorder=0)
        ax_scat.set_xlim([-maxlim, maxlim])
        ax_scat.set_ylim([-maxlim, maxlim])
        ax_scat.set_xticks([])
        ax_scat.set_yticks([])
        ax_scat.set_xlabel("correlation with %s" % labelA, fontsize=fontsize_focus)
        ax_scat.set_ylabel("correlation with %s" % labelB, fontsize=fontsize_focus)

        fra = 0.1 / (len(freqs) - focus)

        con = ConnectionPatch(
            xyA=(maxlim * 0.75, -maxlim),
            coordsA="data",
            xyB=(-maxlim * 0.75, -maxlim),
            coordsB="data",
            axesA=ax_focus,
            axesB=ax_scat,
            arrowstyle="->",
            connectionstyle=f"bar,fraction={fra}",
            **kwargs_con,
        )
        ax_scat.add_artist(con)
        con.set_linewidth(2)
        con.set_color("red")

        fA = pd.concat(
            [corA.loc["genesetA"][str(focus)], corB.loc["genesetA"][str(focus)]], axis=1
        )
        fA = fA.loc[
            (fA.iloc[:, 0].values > 0) & (fA.iloc[:, 1].values < 0),
        ]
        fB = pd.concat(
            [corA.loc["genesetB"][str(focus)], corB.loc["genesetB"][str(focus)]], axis=1
        )
        fB = fB.loc[
            (fB.iloc[:, 1].values > 0) & (fB.iloc[:, 0].values < 0),
        ]
        if top_focus > 0:
            topA = np.flip(
                pairwise_distances(np.array([0, 0]).reshape(1, -1), fA).argsort()
            )[0][:top_focus]
            topB = np.flip(
                pairwise_distances(np.array([0, 0]).reshape(1, -1), fB).argsort()
            )[0][:top_focus]
            texts_A = [
                ax_scat.text(fA.loc[t][0], fA.loc[t][1], t, **kwargs_text)
                for t in fA.index[topA]
            ]
            texts_B = [
                ax_scat.text(fB.loc[t][0], fB.loc[t][1], t, **kwargs_text)
                for t in fB.index[topB]
            ]
            texts = texts_A + texts_B
            adjust_text(texts, **kwargs_adjust)

    savefig_or_show("slide_cors", show=show, save=save)



from typing import Union, Optional
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.colors import hex2color
import seaborn as sns
import igraph
import numpy as np

from scanpy.plotting._utils import savefig_or_show
from typing_extensions import Literal


def module_inclusion(
    adata,
    root_milestone,
    milestones,
    bins: int,
    branch: str,
    figsize: tuple = (6, 5),
    max_t: Literal["fork", "max"] = "max",
    perm: bool = False,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
):
    """\
    Plot results generated from tl.module_inclusion.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    bins
        Number of bins to separate the pseudotime path.
    branch
        Which endpoint to focus on.
    figsize
        Size of the figure.
    max_t
        Until which pseudotime to limit the plot and binning.
    perm
        Show permutation results.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    graph = adata.uns["graph"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    uns_temp = adata.uns.copy()

    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    perm_str = "_perm" if perm else ""
    matSwitch = adata.uns[name]["module_inclusion" + perm_str]

    fork = list(
        set(img.get_shortest_paths(str(root), str(leaves[0]))[0]).intersection(
            img.get_shortest_paths(str(root), str(leaves[1]))[0]
        )
    )
    fork = np.array(img.vs["name"], dtype=int)[fork]
    fork_t = adata.uns["graph"]["pp_info"].loc[fork, "time"].max()

    maxt = fork_t if max_t == "fork" else matSwitch[branch].max().max()
    sg = np.linspace(0, maxt, bins)
    sort = matSwitch[branch].mean(axis=1).sort_values()
    sort = sort[sort < maxt] if max_t == "fork" else sort
    sort = sort[~np.isnan(sort)].index
    matSwitch[branch] = matSwitch[branch].loc[sort]
    hm = np.vstack(
        matSwitch[branch]
        .apply(lambda x: np.array([sum(s > x) / len(x) for s in sg]), axis=1)
        .values
    )

    c_mil = np.array(mlsc)[np.argwhere(adata.obs.milestones.cat.categories == branch)][
        0
    ][0]
    gg = LinearSegmentedColormap.from_list("", ["lightgrey", c_mil])

    # Set up the matplotlib figure
    f, ax = plt.subplots(figsize=figsize)

    sns.heatmap(
        hm,
        yticklabels=sort,
        xticklabels=np.round(sg, 2),
        cmap=gg,
        ax=ax,
        vmin=0,
        vmax=1,
        cbar_kws={"shrink": 0.3, "anchor": (0, 0)},
    )
    ax.set_xlabel("pseudotime")

    ax.axhline(y=0, color="k", linewidth=2)
    ax.axhline(y=hm.shape[0], color="k", linewidth=2)
    ax.axvline(x=0, color="k", linewidth=2)
    ax.axvline(x=hm.shape[1], color="k", linewidth=2)

    if max_t == "max":
        ax.axvline(
            fork_t / (matSwitch[branch].max().max() / bins),
            color="k",
            linestyle="dashed",
        )

    if show == False:
        return axs

    savefig_or_show("module_inclusion", show=show, save=save)



import numpy as np
import pandas as pd
from anndata import AnnData
import igraph

import matplotlib.pyplot as plt
from skmisc.loess import loess
from matplotlib import lines
from typing import Union, Optional

from scanpy.plotting._utils import savefig_or_show

from scFates.tools.utils import importeR

Rpy2, R, rstats, rmgcv, Formula = importeR("Plotting syncho path fit")


def synchro_path(
    adata: AnnData,
    root_milestone,
    milestones,
    knots: int = 10,
    alpha_multi: float = 0.1,
    knots_multi: Union[int, None] = None,
    max_t: Union[float, None] = None,
    figsize: Union[tuple, None] = None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
):
    """\
    Plot results generated from tl.synchro_path.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    knots
        number of knots for gamfit.
    alpha_multi
        alpha multiple mapping.
    knots_multi
        knots for overall mappings (plot disabled if not set).
    max_t
        set max pseudotime to show.
    figsize
        figure size.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    plt.rcParams["axes.grid"] = False

    graph = adata.uns["graph"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    mlsc = np.array(adata.uns["milestones_colors"].copy())
    if isinstance(mlsc, (list)):
        mlsc = np.array(mlsc)
    # mlsc_temp = mlsc.copy()

    dct = graph["milestones"]
    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    fork = list(
        set(img.get_shortest_paths(str(root), str(leaves[0]))[0]).intersection(
            img.get_shortest_paths(str(root), str(leaves[1]))[0]
        )
    )
    fork = np.array(img.vs["name"], dtype=int)[fork]
    fork_t = adata.uns["graph"]["pp_info"].loc[fork, "time"].max()

    allcor = adata.uns[name]["synchro"]
    runs = ["real", "permuted"] if len(list(allcor.keys())) == 2 else ["real"]

    fig, axs = plt.subplots(
        3, len(runs), figsize=(len(runs) * 6, 6) if figsize is None else figsize
    )
    fig.subplots_adjust(hspace=0.05, wspace=0.025, top=0.95)
    i = 0

    dct_cormil = dict(
        zip(
            ["corAA", "corBB", "corAB"],
            [milestones[0] + "\nintra-module", milestones[1] + "\nintra-module"]
            + [milestones[0] + " vs " + milestones[1] + "\ninter-module"],
        )
    )

    axs = axs.ravel(order="F")

    for run in runs:
        for cc in ["corAA", "corBB", "corAB"]:
            for mil in milestones:
                res = allcor[run][mil]
                nmaps = res.n_map.unique()
                col = mlsc[adata.obs.milestones.cat.categories == mil][0]

                if len(nmaps) == 1:
                    m = rmgcv.gam(
                        Formula("%s ~ s(t, bs = 'cs',k=%s)" % (cc, knots)), data=res
                    )
                    pred, se = rmgcv.predict_gam(m, se_fit=True)

                    fit = pred
                    ll = pred - se
                    ul = pred + se

                    axs[i].plot(
                        res.t.values, res[cc], "+", c=col, alpha=0.5, rasterized=True
                    )
                    axs[i].plot(res.t.values, fit, c=col, rasterized=True)
                    axs[i].fill_between(
                        res.t.values.tolist(),
                        ll.tolist(),
                        ul.tolist(),
                        alpha=0.33,
                        edgecolor=col,
                        facecolor=col,
                        rasterized=True,
                    )
                else:
                    for mp in nmaps:
                        m = rmgcv.gam(
                            Formula("%s ~ s(t, bs = 'cs',k=%s)" % (cc, knots)),
                            data=res.loc[res.n_map == mp],
                        )
                        pred = rmgcv.predict_gam(m)
                        axs[i].plot(
                            res.loc[res.n_map == mp].t,
                            res.loc[res.n_map == mp][cc],
                            c=col,
                            alpha=0.01,
                            rasterized=True,
                        )
                        axs[i].plot(
                            res.loc[res.n_map == mp].t,
                            pred,
                            c=col,
                            alpha=alpha_multi,
                            rasterized=True,
                        )

                axs[i].axvline(fork_t, color="black")
                axs[i].axhline(0, linestyle="dashed", color="grey", zorder=0)

                if i == 0:
                    vertical_line = lines.Line2D(
                        [],
                        [],
                        color="black",
                        marker="|",
                        linestyle="None",
                        markersize=10,
                        markeredgewidth=1.5,
                        label="bifurcation",
                    )

                    axs[i].legend(handles=[vertical_line])

                if (i < 2) | ((i > 2) & (i < 5)):
                    axs[i].set_xticks([])
                else:
                    axs[i].set_xlabel("pseudotime")

                if i <= 2:
                    axs[i].set_ylabel(dct_cormil[cc])

                if i > 2:
                    axs[i].set_ylim(axs[i - 3].get_ylim())
                    axs[i].set_yticks([])

                if i == 0:
                    axs[i].set_title("real")

                if i == 3:
                    axs[i].set_title("permuted")

                axs[i].grid(b=None)

            if max_t is not None:
                axs[i].set_xlim([0, max_t])
            i = i + 1

    if show == False:
        return axs

    savefig_or_show("synchro_path", show=show, save=save)



from scFates import logging as logg
from scFates import settings
from scFates.tools.utils import getpath, get_X, ProgressParallel

import igraph
from joblib import delayed
import sys

import numpy as np
import statsmodels.formula.api as smf
from anndata import AnnData
import pandas as pd


def linearity_deviation(
    adata: AnnData,
    start_milestone,
    end_milestone,
    percentiles=[20, 80],
    n_jobs: int = 1,
    n_map: int = 1,
    plot: bool = False,
    basis: str = "X_umap",
    copy: bool = False,
):
    """\
    Identifies genes that specifically characterize a given transition but not the progenitors neither the progenies.

    This approach has been developped in the following study [Kameneva21]_.
    Designed to test whether a bridge/transition is the result of a doublet population only, this test checks if a gene expression occuring in the transition/bridge could be explained by a linear mixture of expressions of that gene in progenitors and progenies. The gene expression profile of each cell of the bridge is modeled as a linear combination of mean gene expression profiles in progenitors and progenies.

    For each gene in each cell in bridge is calculated the magnitude of the residuals not explained by the model. The mean residuals across all cells in the transition/bridge is then normalized to the standard deviation of the expression of a given gene. The obtained normalized mean residual values is used to prioritize the genes with distinctive expression patterns in the bridge population.

    Parameters
    ----------
    adata
        Annotated data matrix.
    start_milestone
        tip defining progenitor branch.
    end_milestone
        tips defining the progeny branch.
    percentiles
        pseudotime percentiles to define the progenitor and progeny populations
    n_jobs
        number of cpu processes used to perform the test.
    n_map
        number of cell mappings from which to do the test.
    plot
        plot the cells selection according to percentiles.
    basis
        basis to use in case of plotting
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.var['A->B_rss']`
            pearson residuals of the linear fit.
        `.obs['A->B_lindev_sel']`
            cell selections used for the test.

    """

    logg.info("Estimation of deviation from linearity", reset=True)

    adata = adata.copy() if copy else adata

    name = start_milestone + "->" + end_milestone

    graph = adata.uns["graph"]

    def lindev_map(m):
        df = adata.uns["pseudotime_list"][str(m)].copy()
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        cells = getpath(
            img,
            graph["milestones"][start_milestone],
            graph["tips"],
            graph["milestones"][end_milestone],
            graph,
            df,
        )

        t_perc = [cells.t.quantile(p / 100) for p in percentiles]

        df[name + "_lindev_sel"] = np.nan
        df.loc[
            adata.obs_names.isin(cells.index), name + "_lindev_sel"
        ] = "putative bridge"
        df.loc[
            (adata.obs_names.isin(cells.index)) & (adata.obs.t < t_perc[0]),
            name + "_lindev_sel",
        ] = "putative progenitors"
        df.loc[
            (adata.obs_names.isin(cells.index)) & (adata.obs.t > t_perc[1]),
            name + "_lindev_sel",
        ] = "putative progenies"

        if m == 0:
            adata.obs[name + "_lindev_sel"] = df[name + "_lindev_sel"]

        progenitors = adata.obs_names[
            df[name + "_lindev_sel"] == "putative progenitors"
        ]
        progenies = adata.obs_names[df[name + "_lindev_sel"] == "putative progenies"]
        bridge = adata.obs_names[df[name + "_lindev_sel"] == "putative bridge"]

        X_all = get_X(adata, cells.index, adata.var_names, layer=None)
        X_progenitors = get_X(adata, progenitors, adata.var_names, layer=None)
        X_progenies = get_X(adata, progenies, adata.var_names, layer=None)
        X_bridge = get_X(adata, bridge, adata.var_names, layer=None)

        X = X_bridge.tolist()

        A = X_progenitors.mean(axis=0)
        B = X_progenies.mean(axis=0)

        def get_resid(x):
            model = smf.ols(
                formula="x ~ A + B - 1", data=pd.DataFrame({"x": x, "A": A, "B": B})
            )
            results = model.fit()
            return results.resid_pearson

        rs = ProgressParallel(
            total=len(X),
            n_jobs=n_jobs,
            use_tqdm=n_map == 1,
            desc="    cells on the bridge",
            file=sys.stdout,
        )(delayed(get_resid)(x) for x in X)

        rs = np.vstack(rs).T
        return rs.mean(axis=1) / X_all.std(axis=0)

    n_jobs_map = 1
    if n_map > 1:
        n_jobs_map = n_jobs
        n_jobs = 1

    rss = ProgressParallel(
        total=n_map,
        n_jobs=n_jobs_map,
        use_tqdm=n_map > 1,
        desc="    multi mapping",
        file=sys.stdout,
    )(delayed(lindev_map)(m) for m in range(n_map))

    if plot:
        import scanpy as sc

        sc.pl.embedding(adata, basis=basis, color=name + "_lindev_sel")

    adata.var[name + "_rss"] = np.nanmean(np.vstack(rss), axis=0)

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .var['" + name + "_rss'], pearson residuals of the linear fit.\n"
        "    .obs['" + name + "_lindev_sel'], cell selections used for the test."
    )

    return adata if copy else None



from typing_extensions import Literal
import numpy as np
import matplotlib.pyplot as plt

from .graph_fitting import *
from .pseudotime import *
from .root import root
from .. import logging as logg
from .. import settings


def cellrank_to_tree(
    adata,
    time,
    Nodes: int,
    method: Literal["ppt", "epg"] = "ppt",
    ppt_lambda=20,
    auto_root=True,
    min_val=False,
    reassign_pseudotime=True,
    copy=False,
    **kwargs
):
    """\
    Converts CellRank [Lange21]_ fate probabilities into a principal tree that can be analysed by scFates.

    It combines the projection generated by `cr.pl.circular_projection`
    with any measure of differentiation (CytoTRACE, latent time). A tree is fitted onto this new embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    time
        time key to use for the additional dimension used in combination with `cr.pl.circular_projection`.
    Nodes
        Number of nodes that compose the principal graph.
    method
        If ppt, uses simpleppt approach, `ppt_lambda` and `ppt_sigma` are the
        parameters controlling the algorithm. If `epg`, uses ComputeElasticPrincipalTree
        function from elpigraph python package, `epg_lambda` `epg_mu` and `epg_trimmingradius`
        are the parameters controlling the algorithm.
    ppt_lambda
        Parameter for simpleppt, penalty for the tree length [Mao15]_. Usually works well at default for the conversion.
    auto_root
        Automatically select the root tip using the time key.
    min_val
        min_val parameter from :func:`scFates.tl.root`
    reassign_pseudotime
        whether use the time key to replace the distances comptued from the tree.
    copy
        Return a copy instead of writing to adata.
    kwargs
        arguments to pass to function :func:`scFates.tl.tree`.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.obsm['X_fates']`
            representation generated by combining the time key with projection generated by :func:`cellrank.pl.circular_projection`.
        `.uns['ppt']`
            dictionnary containing information from simpelppt tree if method='ppt'
        `.uns['epg']`
            dictionnary containing information from elastic principal tree if method='epg'
        `.uns['graph']['B']`
            adjacency matrix of the principal points
        `.uns['graph']['R']`
            soft assignment of cells to principal point in representation space
        `.uns['graph']['F']`
            coordinates of principal points in representation spac


    """
    try:
        import cellrank as cr
    except Exception as e:
        raise Exception(
            'cellrank installation is necessary for conversion. \
            \nPlease use "pip3 install cellrank" to install it'
        )

    logg.info(
        "Converting CellRank results to a principal tree",
        end=" " if settings.verbosity > 2 else "\n",
    )

    adata = adata.copy() if copy else adata

    n_states = adata.obsm["to_terminal_states"].shape[1]

    if n_states == 2:
        adata.obsm["X_fates"] = np.vstack(
            [
                np.array(adata.obsm["to_terminal_states"][:, 0].flatten()),
                adata.obs[time],
            ]
        ).T
        logg.hint(
            "with .obsm['X_fates'], created by combining:\n"
            "    .obsm['to_terminal_states'][:,0] and adata.obs['" + time + "']\n"
        )
    else:
        logg.hint(
            "with .obsm['X_fates'], created by combining:\n"
            "    .obsm['X_fate_simplex_fwd'] (from cr.pl.circular_projection) and adata.obs['"
            + time
            + "']\n"
        )
        cr.pl.circular_projection(adata, keys=["kl_divergence"])
        plt.close()

        adata.obsm["X_fates"] = np.concatenate(
            [adata.obsm["X_fate_simplex_fwd"], adata.obs[time].values.reshape(-1, 1)],
            axis=1,
        )

    tree(
        adata,
        Nodes=Nodes,
        use_rep="X_fates",
        method=method,
        ppt_lambda=ppt_lambda,
        **kwargs
    )

    if auto_root:
        logg.info("\nauto selecting a tip as a root using " + time + ".\n")
        root(adata, time, tips_only=True, min_val=rev_root)
        pseudotime(adata)

    newt = ""
    if reassign_pseudotime:
        adata.obs["t"] = adata.obs[time]

        adata.uns["pseudotime_list"]["0"]["t"] = adata.obs[time]
        for n in range(adata.uns["graph"]["R"].shape[1]):
            adata.uns["graph"]["pp_info"].loc[n, "time"] = np.average(
                adata.obs.t, weights=adata.uns["graph"]["R"][:, n]
            )

        for n in adata.uns["graph"]["pp_seg"].index:
            adata.uns["graph"]["pp_seg"].loc[n, "d"] = np.diff(
                adata.uns["graph"]["pp_info"]
                .loc[adata.uns["graph"]["pp_seg"].loc[n, ["from", "to"]].values, "time"]
                .values
            )[0]

        newt = "    .obs['t'] has been replaced by .obs['" + time + "']\n"

    logg.info("\nfinished", time=True, end="\n" if reassign_pseudotime else " ")
    logg.info(
        "    .obsm['X_fates'] representation used for fitting the tree.\n"
        + newt
        + "    .uns['graph']['pp_info'].time has been updated with "
        + time
        + "\n"
        "    .uns['graph']['pp_seg'].d has been updated with " + time
    )

    return adata if copy else None
